"""Contains the FiniteStateMachine class."""
from collections import namedtuple


__all__ = ["FiniteStateMachine"]


NodeType = namedtuple("NodeType", ["checker", "handler", "state"])


class FiniteStateMachine(object):
  """An implementation of a finite state machine."""
  def __init__(self, graph, initial_node_type, node_handler):
    """Construct the FiniteStateMachine instance.

    :graph: a dictionary of NodeType => Tuple[NodeType]
    :initial_node_type: the value last_node_type should be set to when parsing a new stream
    :node_handler: a function accepting (NodeType.handler(data), NodeType, raw_data)
    """
    self.graph = graph
    self.initial_node_type = initial_node_type
    self.handle = node_handler
    self.last_node_type = None

  def _handle_node(self, data):
    """Given a piece of data, check each node_type that the last_node_type can lead to and act using
    the first whose checker returns true.
    """
    next_options = self.graph[self.last_node_type]
    for node_type in next_options:
      checker, handler, *_ = node_type
      if checker(data):
        self.last_node_type = node_type
        self.handle(handler(data), node_type, data)
        return
    raise Exception("Unable to handle data: {}".format(data))

  def step(self, node):
    """Steps the machine up with a given node."""
    self._handle_node(node)

  def parse(self, nodes):
    """Given a stream of node data, try to parse the nodes according to the machine's graph."""
    self.last_node_type = self.initial_node_type
    for node_number, node in enumerate(nodes):
      try:
        self.step(node)
      except Exception as ex:
        raise Exception("An error occurred on node {}".format(node_number)) from ex



import os
from collections import namedtuple

from bidon.configuration import Configuration
from bidon.db.access import DataAccess, ModelAccess


CONFIG = Configuration(core=None, test_rowcount=False, test_callproc=False, is_pg=False)
TEST_ROOT = os.path.dirname(os.path.abspath(__file__))


def configure(args):
  database = args.database
  if database == "postgres":
    configure_postgres(args)
  elif database == "mysql":
    configure_mysql(args)
  elif database == "sqlite":
    configure_sqlite(args)


def configure_postgres(args):
  from psycopg2.extras import Json
  from psycopg2.extensions import register_adapter

  from bidon.db.core import get_pg_core

  port = args.port or 5432

  CONFIG.update(core=get_pg_core("dbname=bidon_test user=postgres host=localhost port={}".format(port)),
                test_rowcount=True,
                test_callproc=True,
                is_pg=True)
  CONFIG.freeze()

  register_adapter(dict, lambda d: Json(d))


def configure_mysql(args):
  from pymysql.cursors import Cursor

  from bidon.db.core import get_mysql_core

  class NamedTupleCursorMixin(object):
    def _do_get_result(self):
      super(NamedTupleCursorMixin, self)._do_get_result()
      if self.description:
        fields = [f.name for f in self._result.fields]
        Row = namedtuple("Row", fields)
        self._rows = [Row(*row) for row in self._rows]

  class NamedTupleCursor(NamedTupleCursorMixin, Cursor):
    pass

  defaults_file = os.path.join(os.path.expanduser("~"), ".mysql-defaults/localhost")
  CONFIG.update(core=et_mysql_core(dict(read_default_file=defaults_file, database="bidon_test"),
                                   cursor_factory=NamedTupleCursor),
                test_rowcount=True,
                test_callproc=True)
  CONFIG.freeze()


def configure_sqlite(args):
  from bidon.db.core import get_sqlite_core

  def get_namedtuple_factory():
    fields = []
    Row = None
    def namedtuple_factory(cursor, row):
      nonlocal fields, Row
      _fields = [col[0] for col in cursor.description]
      if _fields != fields:
        fields = _fields
        Row = namedtuple("Row", fields)
      return Row(*row)
    return namedtuple_factory

  CONFIG.update(core=get_sqlite_core("tests/fixtures/test.sqlite3",
                                     cursor_factory=get_namedtuple_factory()),
                test_rowcount=False,
                test_callproc=False)
  CONFIG.freeze()


def get_data_access():
  return DataAccess(CONFIG.core)


def get_model_access():
  return ModelAccess(CONFIG.core)


from .db import *
from .util import *
from .spreadsheet import *
from .xml import XMLStreamWriterTestCase
from .test_configuration import ConfigurationTestCase
from .test_data_table import DataTableTestCase
from .test_field_mapping import FieldMappingTestCase
from .test_json_patch import JSONPatchTestCase



import xml.etree.ElementTree as ET
import json
import unittest
from pprint import pprint

import bidon.json_patch as JP
from bidon.util import convert as conv
from bidon.util import transform as T


XML_SOURCE = """<?xml version="1.0" encoding="UTF-8"?>
<ccb_api>
     <request>
          <parameters>
               <argument value="attendance_profiles" name="srv"/>
               <argument value="2014-01-01" name="end_date"/>
               <argument value="2013-01-01" name="start_date"/>
          </parameters>
     </request>
     <response>
          <events count="68">
               <event id="1393">
                    <name>Starbucks Gathering</name>
                    <occurrence>2013-01-21 19:00:00</occurrence>
                    <did_not_meet>false</did_not_meet>
                    <topic></topic>
                    <notes></notes>
                    <prayer_requests></prayer_requests>
                    <info></info>
                    <attendees>
                         <attendee id="327">
                              <first_name>Bob</first_name>
                              <last_name>Absmith</last_name>
                         </attendee>
                         <attendee id="7">
                              <first_name>Amy</first_name>
                              <last_name>Amgar</last_name>
                         </attendee>
                         <attendee id="265">
                              <first_name>Madeline</first_name>
                              <last_name>Ash</last_name>
                         </attendee>
                         <attendee id="227">
                              <first_name>Tasha</first_name>
                              <last_name>Ash</last_name>
                         </attendee>
                         <attendee id="1">
                              <first_name>Thomas</first_name>
                              <last_name>Greenwood</last_name>
                         </attendee>
                         <attendee id="198">
                              <first_name>Rachael</first_name>
                              <last_name>Peaslee-Mueller</last_name>
                         </attendee>
                    </attendees>
                    <head_count></head_count>
               </event>
               <event id="1393">
                    <name>Starbucks Gathering</name>
                    <occurrence>2013-01-21 19:00:00</occurrence>
                    <did_not_meet>false</did_not_meet>
                    <topic></topic>
                    <notes></notes>
                    <prayer_requests></prayer_requests>
                    <info></info>
                    <attendees>
                         <attendee id="327">
                              <first_name>Bob</first_name>
                              <last_name>Absmith</last_name>
                         </attendee>
                         <attendee id="7">
                              <first_name>Amy</first_name>
                              <last_name>Amgar</last_name>
                         </attendee>
                         <attendee id="265">
                              <first_name>Madeline</first_name>
                              <last_name>Ash</last_name>
                         </attendee>
                         <attendee id="227">
                              <first_name>Tasha</first_name>
                              <last_name>Ash</last_name>
                         </attendee>
                         <attendee id="1">
                              <first_name>Thomas</first_name>
                              <last_name>Greenwood</last_name>
                         </attendee>
                         <attendee id="198">
                              <first_name>Rachael</first_name>
                              <last_name>Peaslee-Mueller</last_name>
                         </attendee>
                    </attendees>
                    <head_count></head_count>
               </event>
          </events>
     </response>
</ccb_api>
"""

JSON_SOURCE = """{
  "funds": {
    "fund": [
      {
        "@array": "true",
        "@id": "12462",
        "@oldID": "",
        "@uri": "https://demo.fellowshiponeapi.com/giving/v1/funds/12462",
        "accountReference": {
          "@id": "",
          "@uri": ""
        },
        "createdByPerson": {
          "@id": "2229409",
          "@uri": "{{CONSUMER_ROOT_DOMAIN}}/people/2229409"
        },
        "createdDate": "2010-05-03T11:13:50",
        "fundCode": null,
        "fundType": {
          "@id": "1",
          "@uri": "https://demo.fellowshiponeapi.com/giving/v1/funds/fundtypes/1",
          "name": "Contribution"
        },
        "isActive": "true",
        "isWebEnabled": "true",
        "lastUpdatedByPerson": {
          "@id": "2229409",
          "@uri": "{{CONSUMER_ROOT_DOMAIN}}/people/2229409"
        },
        "lastUpdatedDate": "2010-05-27T11:41:23",
        "name": "Brotherhood"
      },
      {
        "@array": "true",
        "@id": "4895",
        "@oldID": "",
        "@uri": "https://demo.fellowshiponeapi.com/giving/v1/funds/4895",
        "accountReference": {
          "@id": "",
          "@uri": ""
        },
        "createdByPerson": {
          "@id": "2229409",
          "@uri": "{{CONSUMER_ROOT_DOMAIN}}/people/2229409"
        },
        "createdDate": "2010-05-03T11:13:50",
        "fundCode": null,
        "fundType": {
          "@id": "1",
          "@uri": "https://demo.fellowshiponeapi.com/giving/v1/funds/fundtypes/1",
          "name": "Contribution"
        },
        "isActive": "false",
        "isWebEnabled": "false",
        "lastUpdatedByPerson": {
          "@id": "2229409",
          "@uri": "{{CONSUMER_ROOT_DOMAIN}}/people/2229409"
        },
        "lastUpdatedDate": "2010-05-27T11:41:23",
        "name": "Building Campaign"
      }
    ]
  }
}
"""

class UtilTransformTestCase(unittest.TestCase):
  def test_xml(self):
    def to_bool(val):
      return val == "true"

    root = ET.fromstring(XML_SOURCE)
    events = root.find("response/events")

    event_transform = dict(
      id=T.val(T.xml_attr("id"), conv.to_int),
      name=T.val(T.xml_text("name")),
      date=T.val(T.xml_text("occurrence"), conv.to_date),
      not_held=T.val(T.xml_text("did_not_meet"), to_bool),
      head_count=T.val(T.xml_text("head_count"), conv.accept_none_wrapper(conv.to_int)),
      attendees=T.lst(T.xml_children("attendees/attendee"), T.obj(None, dict(
        id=T.val(T.xml_attr("id"), conv.to_int),
        first_name=T.val(T.xml_text("first_name")),
        last_name=T.val(T.xml_text("last_name"))
      )))
    )

    l = T.get_lst(events, T.xml_children("event"), T.obj(None, event_transform))
    self.assertEqual(len(l), 2)
    e = l[0]
    self.assertEqual(set(e.keys()), {"attendees", "id", "date", "head_count", "not_held", "name"})
    al = e["attendees"]
    self.assertEqual(len(al), 6)
    a = al[0]
    self.assertEqual(set(a.keys()), {"id", "first_name", "last_name"})

  def test_json(self):
    def to_bool(val):
      return val == "true"

    root = json.loads(JSON_SOURCE)

    fund_transform = dict(
      id=T.val(T.json_val("/@id"), conv.to_int),
      created_date=T.val(T.json_val("/createdDate"), conv.to_datetime),
      created_by_user_id=T.val(T.json_val("/createdByPerson/@id"), conv.to_int),
      fund_type=T.obj(T.json_val("/fundType"), dict(
        id=T.val(T.json_val("/@id"), conv.to_int),
        name=T.val(T.json_val("/name")))),
      is_active=T.val(T.json_val("/isActive"), to_bool)
    )

    l = T.get_lst(root, T.json_vals("/funds/fund/*"), T.obj(None, fund_transform, T.flatten({"fund_type"})))
    self.assertEqual(len(l), 2)
    f = l[0]
    self.assertEqual(set(f.keys()), {"created_by_user_id", "created_date", "fund_type_id", "fund_type_name", "id", "is_active"})
    # pprint(l)

  def test_json_bad_path(self):
    d = dict(a=dict(b=1, c=2))
    self.assertIsNone(T.get_json_val(d, "/a/d", ignore_bad_path=True))
    with self.assertRaises(JP.JSONPathError):
      T.get_json_val(d, "/a/d")



"""The sql_writer module contains the SqlWriter class."""
import operator
import re


__all__ = ["SqlWriter"]


class SqlWriter(object):
  """This class writes SQL format strings, it can be customized for different driver conventions for
  placeholders and different typecaster conventions for databases.
  """
  def __init__(self, named_placeholder, unnamed_placeholder, typecaster, *, sort_columns=False):
    """Construct the instance.

    :named_placeholder: must be a format string accepting a single item.
    :unnamed_placeholder: must be a string.
    :typecaster: must be a format string accepting two items: a subject and a db type.
    """
    self.named_placeholder = named_placeholder
    self.unnamed_placeholder = unnamed_placeholder
    self.typecaster = typecaster
    self.sort_columns = sort_columns

  def typecast(self, subject, db_type):
    """Applys subject and db_type to the instance's typecaster format string."""
    return self.typecaster.format(subject, db_type)

  def to_placeholder(self, name=None, db_type=None):
    """Returns a placeholder for the specified name, by applying the instance's format strings.

    :name: if None an unamed placeholder is returned, otherwise a named placeholder is returned.
    :db_type: if not None the placeholder is typecast.
    """
    if name is None:
      placeholder = self.unnamed_placeholder
    else:
      placeholder = self.named_placeholder.format(name)

    if db_type:
      return self.typecast(placeholder, db_type)
    else:
      return placeholder

  def to_tuple(self, iterable, surround="()", joiner=", "):
    """Returns the iterable as a SQL tuple."""
    return "{0}{1}{2}".format(surround[0], joiner.join(iterable), surround[1])

  def to_expression(self, lhs, rhs, op):
    """Builds a binary sql expression.

    At its most basic, returns 'lhs op rhs' such as '5 + 3'. However, it also specially handles the
    'in' and 'between' operators. For each of these operators it is expected that rhs will be
    iterable.

    If the comparison operator is of the form 'not(op)' where op is the operator, it will result in
    not (lhs op rhs).

    This allows for doing the full range of null checks on composite types. For composite types,
    'is null' only returns true when all fields are null, and 'is not null' returns true only when
    all fields are not null. So, for a composite type with some null fields, 'is null' and
    'is not null' will both return false, making it difficult to get all rows that have composite
    columns with some value in them. The solution to this is to use not (composite is null) which
    is true when all composite fields, or only some composite fields are not null.
    """
    if op == "raw":
      # TODO: This is not documented
      return lhs
    elif op == "between":
      return "{0} between {1} and {2}".format(lhs, *rhs)
    elif op == "in":
      return "{0} in {1}".format(lhs, self.to_tuple(rhs))
    elif op.startswith("not(") and op.endswith(")"):
      return "not ({0} {1} {2})".format(lhs, op[4:-1].strip(), rhs)
    else:
      return "{0} {1} {2}".format(lhs, op.strip(), rhs)

  def transform_op(self, op, value):
    """For comparisons, if the value is None (null), the '=' operator must be replaced with ' is '
    and the '!=' operator must be replaced with ' is not '. This function handles that conversion.
    It's up to the caller to call this function only on comparisons and not on assignments.
    """
    if value is None:
      if _EQ_RE.match(op):
        return "is"
      elif _NEQ_RE.match(op):
        return "is not"

    return op

  def value_comparisons(self, values, comp="=", is_assignment=False):
    """Builds out a series of value comparisions.

    :values: can either be a dictionary, in which case the return will compare a name to a named
    placeholder, using the comp argument. I.E. values = {"first_name": "John", "last_name": "Smith"}
    will return ["first_name = %(first_name)s", "last_name = %(last_name)s"].

    Otherwise values will be assumed to be an iterable of 2- or 3-tuples in the form
    (column, value[, operator]). When operator is not specified, it will fallback to comp. So for
    instance values = [("first_name", "John"), ("id", (10, 100), "between")] will return
    ["first_name = %s", "id between %s and %s "].

    :is_assigment: if False, transform_op will be called on each operator.
    """
    if isinstance(values, dict):
      if self.sort_columns:
        keys = sorted(values.keys())
      else:
        keys = list(values.keys())

      params = zip(keys, [self.to_placeholder(k) for k in keys])
      return [
        self.to_expression(
          i[0],
          i[1],
          comp if is_assignment else self.transform_op(comp, values[i[0]]))
        for i in params]
    else:
      if self.sort_columns:
        values = sorted(values, key=operator.itemgetter(0))
      comps = []
      for val in values:
        lhs = val[0]
        op = val[2] if len(val) == 3 else comp
        if op == "raw":
          rhs = None
        elif op == "between":
          rhs = (self.to_placeholder(), self.to_placeholder())
        elif op == "in":
          rhs = [self.to_placeholder() for i in val[1]]
        else:
          rhs = self.to_placeholder()
          if not is_assignment:
            op = self.transform_op(op, val[1])

        comps.append(self.to_expression(lhs, rhs, op))
      return comps

  def join_comparisons(self, values, joiner, *, is_assignment=False, comp="="):
    """Generates comparisons with the value_comparisions method, and joins them with joiner.

    :is_assignment: if false, transform_op will be called on each comparison operator.
    """
    if isinstance(values, str):
      return values
    else:
      return joiner.join(self.value_comparisons(values, comp, is_assignment))

  def parse_constraints(self, constraints, joiner=" and ", *, is_assignment=False, comp="="):
    """Parses constraints into a (sql, params) tuple.

    :constraints: can either be a dict, an enumerable with a sql string and params, an enumerable of
    2- or 3-tuples, or just a string.

    If is_assignment is false, transform_op will be called on each comparison operator.
    """
    if constraints is None:
      return (None, None)
    elif isinstance(constraints, dict):
      return (
        self.join_comparisons(constraints, joiner, is_assignment=is_assignment, comp=comp),
        self.get_params(constraints))
    elif isinstance(constraints, (list, tuple)):
      if len(constraints) == 2 and isinstance(constraints[0], str):
        return (constraints[0], constraints[1])
      else:
        return (
          self.join_comparisons(constraints, joiner, is_assignment=is_assignment, comp=comp),
          self.get_params(constraints))
    elif isinstance(constraints, str):
      return (constraints, None)
    else:
      raise TypeError(
        "constraints must be None, str, dict, list or tuple. Is {0}".format(
          type(constraints).__name__))

  def get_params(self, values):
    """Gets params to be passed to execute from values.

    :values: can either be a dict, in which case it will be returned as is, or can be an enumerable
    of 2- or 3-tuples. This will return an enumerable of the 2nd values, and in the case of some
    operators such as 'in' and 'between' the values will be specially handled.
    """
    if values is None:
      return None
    elif isinstance(values, dict):
      return values
    elif isinstance(values, (list, tuple)):
      params = []
      for val in values:
        if len(val) == 2:
          params.append(val[1])
        else:
          if val[2] in ("in", "between"):
            params.extend(val[1])
          else:
            params.append(val[1])
      return params
    elif isinstance(values, str):
      return None
    else:
      raise TypeError(
        "values must be None, a dict, list or tuple, is {0}".format(type(values).__name__))

  def get_find_all_query(self, table_name, constraints=None, *, columns=None, order_by=None,
                         limiting=(None, None)):
    """Builds a find query.

    :limiting: if present must be a 2-tuple of (limit, offset) either of which can be None.
    """
    where, params = self.parse_constraints(constraints)

    if columns:
      if isinstance(columns, str):
        pass
      else:
        columns = ", ".join(columns)
    else:
      columns = "*"

    if order_by:
      order = " order by {0}".format(order_by)
    else:
      order = ""

    paging = ""
    if limiting is not None:
      limit, offset = limiting
      if limit is not None:
        paging += " limit {0}".format(limit)
      if offset is not None:
        paging += " offset {0}".format(offset)

    return ("select {0} from {1} where {2}{3}{4}".format(
      columns,
      table_name,
      where or "1 = 1",
      order,
      paging
    ), params)


_EQ_RE = re.compile(r"^\s*=\s*$")
_NEQ_RE = re.compile(r"^\s*!=\s*$")



"""Contains the TransferTracker class and related classes and functions."""
from bidon.db.access import ModelAccess, ModelBase
from bidon.util import utc_now


__all__ = ["TransferTracker"]


class Relation(ModelBase):
  """A model that keeps track of table transfer completions."""
  table_name = "relations"
  timestamps = None
  attrs = dict(name=None, completed_at=None)


class Transfer(ModelBase):
  """A model that relates the old ID of a moved model to its new ID."""
  table_name = "transfers"
  timestamps = None
  primary_key_name = ("relation_id", "old_id")
  primary_key_is_auto = False
  attrs = dict(relation_id=None, old_id=None, new_id=None)


class TransferTracker(object):
  """The TransferTracker class keeps track of data being transferred from one database or table to
  another. Specifically, it makes it possible to record the old and new ids for a relation, and
  query for the new ID.
  """
  def __init__(self, data_access_core):
    """Initialize the TransferTracker class."""
    self.data_access = ModelAccess(data_access_core)

  def open(self, autocommit=False):
    """Call-through to data_access.open."""
    self.data_access.open(autocommit=autocommit)
    return self

  def close(self, commit=True):
    """Call-through to data_access.close."""
    self.data_access.close(commit=commit)
    return self

  def commit(self):
    """Call-through to data_access.commit."""
    self.data_access.commit()
    return self

  def rollback(self):
    """Call-through to data_access.rollback."""
    self.data_access.rollback()
    return self

  def setup(self):
    """Creates the default relations and transfers tables. The SQL used may not work on all
    databases. (It was written for SQLite3)
    """
    cmds = [
      """
        create table if not exists relations (
          id integer not null primary key,
          name text not null unique,
          completed_at datetime
        );
      """,
      """
        create table if not exists transfers (
          relation_id integer not null references relations (id) on delete cascade,
          old_id text not null,
          new_id text,
          primary key (relation_id, old_id)
        );
      """,
      """
        create index if not exists transfers_relation_id_idx on transfers (relation_id);
      """
    ]

    for cmd in cmds:
      self.data_access.execute(cmd)
    self.data_access.commit()
    return self

  def cleanup(self):
    """Cleanup the database, calling vacuum and analyze."""
    self.data_access.execute("vacuum")
    self.data_access.execute("analyze")

  def reset(self, relation_name=None):
    """Reset the transfer info for a particular relation, or if none is given, for all relations.
    """
    if relation_name is not None:
      self.data_access.delete("relations", dict(name=relation_name))
    else:
      self.data_access.delete("relations", "1=1")
    return self

  def start_transfer(self, relation_name):
    """Write records to the data source indicating that a transfer has been started for a particular
    relation.
    """
    self.reset(relation_name)
    relation = Relation(name=relation_name)
    self.data_access.insert_model(relation)
    return relation

  def register_transfer(self, relation, old_id, new_id):
    """Register the old and new ids for a particular record in a relation."""
    transfer = Transfer(relation_id=relation.id, old_id=old_id, new_id=new_id)
    self.data_access.insert_model(transfer)
    return transfer

  def complete_transfer(self, relation, cleanup=True):
    """Write records to the data source indicating that a transfer has been completed for a
    particular relation.
    """
    relation.completed_at = utc_now().isoformat()
    self.data_access.update_model(relation)
    if cleanup:
      self.cleanup()
    return relation

  def is_transfer_complete(self, relation_name):
    """Checks to see if a tansfer has been completed."""
    phold = self.data_access.sql_writer.to_placeholder()
    return self.data_access.find_model(
      Relation,
      ("name = {0} and completed_at is not null".format(phold), [relation_name])) is not None

  def get_new_id(self, relation_name, old_id, strict=False):
    """Given a relation name and its old ID, get the new ID for a relation. If strict is true, an
    error is thrown if no record is found for the relation and old ID.
    """
    record = self.data_access.find(
      "relations as r inner join transfers as t on r.id = t.relation_id",
      (("r.name", relation_name), ("t.old_id", old_id)),
      columns="new_id")
    if record:
      return record[0]
    else:
      if strict:
        raise KeyError("{0} with id {1} not found".format(relation_name, old_id))

  def id_getter(self, relation_name, strict=False):
    """Returns a function that accepts an old_id and returns the new ID for the enclosed relation
    name."""
    def get_id(old_id):
      """Get the new ID for the enclosed relation, given an old ID."""
      return self.get_new_id(relation_name, old_id, strict)
    return get_id



"""The foreign_model_wrapper module contains the ForeignModelWrapper class."""
from bidon.util.transform import get_obj
from .model_base import ModelBase


__all__ = ["ForeignModelWrapper"]


class ForeignModelWrapper(ModelBase):
  """This class is a specialization on ModelBase that can be configurued with information on how to
  transform a data structure, such a dictionary or an XML tree, to a Model.
  """
  timestamps = None
  table_name = None
  primary_key_name = None
  primary_key_is_auto = False
  transform_args = None

  @classmethod
  def create(cls, source, *, transform_args=None):
    """Create an instance of the class from the source. By default cls.transform_args is used, but
    can be overridden by passing in transform_args.
    """
    if transform_args is None:
      transform_args = cls.transform_args

    return cls(get_obj(source, *transform_args))

  @classmethod
  def map(cls, sources, *, transform_args=None):
    """Generates instances from the sources using either cls.transform_args or transform_args
    argument if present.
    """
    for idx, source in enumerate(sources):
      try:
        yield cls.create(source, transform_args=transform_args)
      except Exception as ex:
        raise Exception("An error occurred with item {0}".format(idx)) from ex




"""This module contains classes and methods that are used to model database tables."""
from .foreign_model_wrapper import ForeignModelWrapper
from .model_base import ModelBase
from .validation import Validation, Validator



"""A collection of utility methods used by various parts of the bidon package."""
import json
import re
import sys
from contextlib import contextmanager
from uuid import UUID

from . import convert
from . import date

_JSON_DEFAULTS = []


# pylint: disable=invalid-name
try_parse_int = convert.try_wrapper(convert.to_int)
try_parse_float = convert.try_wrapper(convert.to_float)
try_parse_decimal = convert.try_wrapper(convert.to_decimal)
try_parse_date = convert.try_wrapper(date.parse_date)
try_parse_time = convert.try_wrapper(date.parse_time)
try_parse_datetime = convert.try_wrapper(date.parse_datetime)
# pylint: enable=invalid-name


IS_UUID_RE = re.compile(r"^[a-z0-9]{8}(-[a-z0-9]{4}){3}-[a-z0-9]{12}$")
def is_uuid(val):
  """Checks whether the value is either an instance of UUID, or if it matches a uuid regex.

  :val: the value to check"""
  return val and (isinstance(val, UUID) or IS_UUID_RE.match(val) != None)


def exclude(source, keys, *, transform=None):
  """Returns a dictionary excluding keys from a source dictionary.

  :source: a dictionary
  :keys: a set of keys, or a predicate function that accepting a key
  :transform: a function that transforms the values
  """
  check = keys if callable(keys) else lambda key: key in keys
  return {key: transform(source[key]) if transform else source[key]
          for key in source if not check(key)}


def pick(source, keys, *, transform=None):
  """Returns a dictionary including only specified keys from a source dictionary.

  :source: a dictionary
  :keys: a set of keys, or a predicate function that accepting a key
  :transform: a function that transforms the values
  """
  check = keys if callable(keys) else lambda key: key in keys
  return {key: transform(source[key]) if transform else source[key]
          for key in source if check(key)}


def register_json_default(pred, convert_):
  """Register a predicate and converter function for json defaults.

  The registered functions are searched in order they are registered.

  :pred: a function that returns true when a particular value should be converted using :convert_:
  :convert_: a function that transforms a value to one that can be included in JSON
  """
  _JSON_DEFAULTS.append((pred, convert_))


def json_default(obj):
  """Convert an object to JSON, via the defaults set with register_json_default.

  :obj: the object to convert
  """
  for default in _JSON_DEFAULTS:
    if default[0](obj):
      return default[1](obj)
  raise TypeError(repr(obj) + " is not JSON serializable")


def to_json(obj, pretty=False):
  """Converts an object to JSON, using the defaults specified in register_json_default.

  :obj: the object to convert to JSON
  :pretty: if True, extra whitespace is added to make the output easier to read
  """
  sort_keys = False
  indent = None
  separators = (",", ":")

  if isinstance(pretty, tuple):
    sort_keys, indent, separators = pretty
  elif pretty is True:
    sort_keys = True
    indent = 2
    separators = (", ", ": ")

  return json.dumps(obj, sort_keys=sort_keys, indent=indent, separators=separators,
                    default=json_default)


def has_value(obj, name):
  """A flexible method for getting values from objects by name.

  returns:
  - obj is None: (False, None)
  - obj is dict: (name in obj, obj.get(name))
  - obj hasattr(name): (True, getattr(obj, name))
  - else: (False, None)

  :obj: the object to pull values from
  :name: the name to use when getting the value
  """
  if obj is None:
    return (False, None)
  elif isinstance(obj, dict):
    return (name in obj, obj.get(name))
  elif hasattr(obj, name):
    return (True, getattr(obj, name))
  elif hasattr(obj, "__getitem__") and hasattr(obj, "__contains__") and name in obj:
    return (True, obj[name])
  else:
    return (False, None)


def get_value(obj, name, fallback=None):
  """Calls through to has_value. If has_value[0] is True, return has_value[1] otherwise returns
  fallback() if fallback is callable, else just fallback.

  :obj: the object to pull values from
  :name: the name to use when getting the value
  :fallback: the value to return when has_value(:obj:, :name:) returns False
  """
  present, value = has_value(obj, name)
  if present:
    return value
  else:
    if callable(fallback):
      return fallback()
    else:
      return fallback


def set_value(obj, name, value):
  """A flexible method for setting a value on an object.

  If the object implements __setitem__ (such as a dict) performs obj[name] = value, else performs
  setattr(obj, name, value).

  :obj: the object to set the value on
  :name: the name to assign the value to
  :value: the value to assign
  """
  if hasattr(obj, "__setitem__"):
    obj[name] = value
  else:
    setattr(obj, name, value)


def with_defaults(method, nparams, defaults=None):
  """Call method with nparams positional parameters, all non-specified defaults are passed None.

  :method: the method to call
  :nparams: the number of parameters the function expects
  :defaults: the default values to pass in for the last len(defaults) params
  """
  args = [None] * nparams if not defaults else defaults + max(nparams - len(defaults), 0) * [None]
  return method(*args)


def namedtuple_with_defaults(ntup, defaults=None):
  """Wraps with_defaults for a named tuple.

  :ntup: the namedtuple constructor
  :defaults: the defaultvalues to pass in for the last len(defaults) params
  """
  return with_defaults(ntup, len(ntup._fields), defaults)


def delegate(from_owner, to_owner, methods):
  """Creates methods on from_owner to call through to methods on to_owner.

  :from_owner: the object to delegate to
  :to_owner: the owner on which to delegate from
  :methods: a list of methods to delegate
  """
  for method in methods:
    _delegate(from_owner, to_owner, method)


def _delegate(from_owner, to_owner, method):
  """Creates a method on from_owner to calls through to the same method on to_owner.

  :from_owner: the object to delegate to
  :to_owner: the owner on which to delegate from
  :methods: the method to delegate
  """
  dgate = lambda self, *args, **kwargs: getattr(getattr(self, to_owner), method)(*args, **kwargs)
  dgate.__name__ = method
  dgate.__doc__ = "Delegates to {0}.{1}: {2}".format(to_owner, method, method.__doc__)
  setattr(from_owner, method, dgate)


def flatten_dict(source, ancestors=None):
  """Flattens a dictionary into (key, value) tuples. Where key is a tuple of ancestor keys.

  :source: the root dictionary
  :ancestors: the tuple of ancestors for every key in :source:"""
  if not ancestors:
    ancestors = ()
  for key in source:
    if isinstance(source[key], dict):
      yield from flatten_dict(source[key], ancestors + (key, ))
    else:
      yield (ancestors + (key, ), source[key])


@contextmanager
def get_file_object(filename, mode="r"):
  """Context manager for a file object. If filename is present, this is the
  same as with open(filename, mode): ...

  If filename is not present, then the file object returned is either
  sys.stdin or sys.stdout depending on the mode.

  :filename: the name of the file, or None for STDIN
  :mode: the mode to open the file with
  """
  if filename is None:
    if mode.startswith("r"):
      yield sys.stdin
    else:
      yield sys.stdout
  else:
    with open(filename, mode) as fobj:
      yield fobj



"""The data_access module contains the DataAccess class and some methods for managing transactions.
"""
import logging
from contextlib import contextmanager

from bidon.util import get_value


__all__ = ["DataAccess", "RollbackTransaction", "Upsert", "transaction", "autocommit"]


logger = logging.getLogger(__name__)


class DataAccess(object):
  """A thin wrapper over a DB API2 database connection.

  Provides a context manager, direct access to the execute method, and wrappers over some common
  methods to reduce the amount of boilerplate SQL required.

  This can work with several database engines, such as MySQL and SQLite, but is primarily targeted
  at Postgres, and includes functionality specific to Postgres features.
  """
  def __init__(self, core, *, search_path=None):
    """Initialize a DataAccess object with a DataAccessCore

    :param core: a DataAccessCore instance
    :param search_path: Postgres only. a list of schema names, which is the order in
                        which to search for unqualified objects.
    """
    self.core = core
    self.connection = None
    self.last_query = None
    self.row_count = -1
    self._search_path = search_path

  def __enter__(self):
    """Opens DataAccess and returns self as context manager."""
    return self.open()

  def __exit__(self, ex_type, ex_value, ex_traceback):
    """Closes DataAccess, requesting commit if no execption was thrown.

    :param ex_type: the type of the exception
    :param ex_value: the parameters passed to the exception
    :param ex_traceback: the traceback object attached to the exception
    """
    self.close(commit=ex_type is None)

  def _update_cursor_stats(self, cursor):
    """Sets DataAccess members according to the given cursor.

    :param cursor: a cursor object
    """
    self.row_count = cursor.rowcount

  @property
  def sql_writer(self):
    """Returns the SqlWriter object associated with the core."""
    return self.core.sql_writer

  def commit(self):
    """Commit the pending statements."""
    logger.debug("Commit")
    self.connection.commit()

  def rollback(self):
    """Rollback the pending statements."""
    logger.debug("Rollback")
    self.connection.rollback()

  @property
  def autocommit(self):
    """Returns the current autocommit state"""
    return self.core.get_autocommit(self.connection)

  @autocommit.setter
  def autocommit(self, value):
    """Set the autocommit value.

    :param value: the new autocommit value
    """
    logger.debug("Setting autocommit from %s to %s", self.autocommit, value)
    self.core.set_autocommit(self.connection, value)

  def _configure_connection(self, name, value):
    """Sets a Postgres run-time connection configuration parameter.

    :param name: the name of the parameter
    :param value: a list of values matching the placeholders
    """
    self.update("pg_settings", dict(setting=value), dict(name=name))

  def open(self, *, autocommit=False):
    """Sets the connection with the core's open method.

    :param autocommit: the default autocommit state
    :type autocommit: boolean
    :return: self
    """
    if self.connection is not None:
      raise Exception("Connection already set")
    self.connection = self.core.open()
    if self._search_path:
      self._configure_connection(
        "search_path",
        self._search_path)
    self.autocommit = autocommit
    return self

  def close(self, *, commit=True):
    """Closes the connection via the core's close method.

    :param commit: if true the current transaction is commited, otherwise it is rolled back
    :type commit: boolean
    :return: self
    """
    self.core.close(self.connection, commit=commit)
    self.connection = None
    return self

  def execute(self, query_string, params=None):
    """Executes a query. Returns the resulting cursor.

    :query_string: the parameterized query string
    :params: can be either a tuple or a dictionary, and must match the parameterization style of the
             query
    :return: a cursor object
    """
    cr = self.connection.cursor()
    logger.info("SQL: %s (%s)", query_string, params)
    self.last_query = (query_string, params)
    cr.execute(query_string, params or self.core.empty_params)
    self._update_cursor_stats(cr)
    return cr

  def callproc(self, name, params, param_types=None):
    """Calls a procedure.

    :param name: the name of the procedure
    :param params: a list or tuple of parameters to pass to the procedure.
    :param param_types: a list or tuple of type names. If given, each param will be cast via
                        sql_writers typecast method. This is useful to disambiguate procedure calls
                        when several parameters are null and therefore cause overload resoluation
                        issues.
    :return: a 2-tuple of (cursor, params)
    """

    if param_types:
      placeholders = [self.sql_writer.typecast(self.sql_writer.to_placeholder(), t)
                      for t in param_types]
    else:
      placeholders = [self.sql_writer.to_placeholder() for p in params]

    # TODO: This may be Postgres specific...
    qs = "select * from {0}({1});".format(name, ", ".join(placeholders))
    return self.execute(qs, params), params

  def get_callproc_signature(self, name, param_types):
    """Returns a procedure's signature from the name and list of types.

    :name: the name of the procedure
    :params: can be either strings, or 2-tuples. 2-tuples must be of the form (name, db_type).
    :return: the procedure's signature
    """
    if isinstance(param_types[0], (list, tuple)):
      params = [self.sql_writer.to_placeholder(*pt) for pt in param_types]
    else:
      params = [self.sql_writer.to_placeholder(None, pt) for pt in param_types]

    return name + self.sql_writer.to_tuple(params)

  def find(self, table_name, constraints=None, *, columns=None, order_by=None):
    """Returns the first record that matches the given criteria.

    :table_name: the name of the table to search on
    :constraints: is any construct that can be parsed by SqlWriter.parse_constraints.
    :columns: either a string or a list of column names
    :order_by: the order by clause
    """
    query_string, params = self.sql_writer.get_find_all_query(
      table_name, constraints, columns=columns, order_by=order_by)
    query_string += " limit 1;"
    return self.execute(query_string, params).fetchone()

  def find_all(self, table_name, constraints=None, *, columns=None, order_by=None, limiting=None):
    """Returns all records that match a given criteria.

    :table_name: the name of the table to search on
    :constraints: is any construct that can be parsed by SqlWriter.parse_constraints.
    :columns: either a string or a list of column names
    :order_by: the order by clause
    """
    query_string, params = self.sql_writer.get_find_all_query(
      table_name, constraints, columns=columns, order_by=order_by, limiting=limiting)
    query_string += ";"
    return self.execute(query_string, params)

  def page(self, table_name, paging, constraints=None, *, columns=None, order_by=None,
           get_count=True):
    """Performs a find_all method with paging.

    :param table_name: the name of the table to search on
    :param paging: is a tuple containing (page, page_size).
    :param constraints: is any construct that can be parsed by SqlWriter.parse_constraints.
    :param columns: either a string or a list of column names
    :param order_by: the order by clause
    :param get_count: if True, the total number of records that would be included without paging are
                      returned. If False, None is returned for the count.
    :return: a 2-tuple of (records, total_count)
    """
    if get_count:
      count = self.count(table_name, constraints)
    else:
      count = None

    page, page_size = paging

    limiting = None
    if page_size > 0:
      limiting = (page_size, page * page_size)

    records = list(self.find_all(
      table_name, constraints, columns=columns, order_by=order_by, limiting=limiting))
    return (records, count)

  def update(self, table_name, values, constraints=None, *, returning=None):
    """Builds and executes and update statement.

    :param table_name: the name of the table to update
    :param values: can be either a dict or an enuerable of 2-tuples in the form (column, value).
    :param constraints: can be any construct that can be parsed by SqlWriter.parse_constraints.
                        However, you cannot mix tuples and dicts between values and constraints.
    :param returning: the columns to return after updating. Only works for cores that support the
                      returning syntax
    :return: a cursor object
    """
    if constraints is None:
      constraints = "1=1"
    assignments, assignment_params = self.sql_writer.parse_constraints(
      values, ", ", is_assignment=True)
    where, where_params = self.sql_writer.parse_constraints(constraints, " and ")
    returns = ""
    if returning and self.core.supports_returning_syntax:
      returns = " returning {0}".format(returning)
    sql = "update {0} set {1} where {2}{3};".format(table_name, assignments, where, returns)
    params = assignment_params

    if constraints is None or isinstance(constraints, str):
      pass
    elif isinstance(constraints, dict):
      if isinstance(params, list):
        raise ValueError("you cannot mix enumerable and dict values and constraints")
      params = params or {}
      params.update(where_params)
    else:
      if isinstance(params, dict):
        raise ValueError("you cannot mix enumerable and dict values and constraints")
      params = params or []
      params.extend(where_params)

    cr = self.execute(sql, params)
    return cr

  def insert(self, table_name, values, *, returning=None, upsert=None):
    """Builds and executes an insert statement.

    :param table_name: the name of the table to insert into
    :param values: can be either a dict or an enumerable of 2-tuples in the form (column, value).
    :param returning: the columns to return after updating. Only works for cores that support the
                      returning syntax
    :param upsert: an Upsert instance, defining how to perform the uspert.
    :return: a cursor object
    """
    if isinstance(values, dict):
      names = values.keys()
      placeholders = [self.sql_writer.to_placeholder(i) for i in names]
      params = values
    else:
      names = [i[0] for i in values]
      placeholders = [self.sql_writer.to_placeholder() for i in values]
      params = [i[1] for i in values]
    placeholders = self.sql_writer.to_tuple(placeholders)
    names = self.sql_writer.to_tuple(names)
    returns = ""
    if returning and self.core.supports_returning_syntax:
      returns = " returning {0}".format(returning)

    if upsert:
      if upsert.action == Upsert.DO_NOTHING:
        u_action_fmt = " on conflict{} do nothing"
        u_assignments = None
        u_constraints = None
      else:
        u_action_fmt = " on conflict{} do update set {}{}"
        u_assignments, u_assignment_params = self.sql_writer.parse_constraints(
          values, ", ", is_assignment=True)
        u_constraints = ""

        if isinstance(values, dict):
          if not upsert.force:
            # Build constraints based on the full column name, and using !=. This will make it so
            # that the record won't be updated if all the fields we care about are the same.
            u_constraints, u_params = self.core.sql_writer.parse_constraints(
              {"{}.{}".format(table_name, k): v for k, v in values.items()},
              " or ",
              comp="!=")
            params.update(u_params)
        else:
          # Add assignment params
          params.extend(u_assignment_params)

          if not upsert.force:
            # Build constraints based on the full column name, and using !=. This will make it so
            # that the record won't be updated if all the fields we care about are the same.
            u_constraints, u_params = self.core.sql_writer.parse_constraints(
              [("{}.{}".format(table_name, col), val, "!=") for (col, val, *_) in values],
              " or ")
            params.extend(u_params)

      if u_constraints:
        u_constraints = " where {}".format(u_constraints)

      u_action_str = u_action_fmt.format(upsert.target_str or "", u_assignments, u_constraints)

      sql = "insert into {0} {1} values {2}{3}{4}".format(
        table_name,
        names,
        placeholders,
        u_action_str,
        returns)
    else:
      sql = "insert into {0} {1} values {2}{3};".format(table_name, names, placeholders, returns)

    cr = self.execute(sql, params)
    return cr

  def delete(self, table_name, constraints=None):
    """Builds and executes an delete statement.

    :param table_name: the name of the table to delete from
    :param constraints: can be any construct that can be parsed by SqlWriter.parse_constraints.
    :return: a cursor object
    """
    if constraints is None:
      constraints = "1=1"
    where, params = self.sql_writer.parse_constraints(constraints)
    sql = "delete from {0} where {1};".format(table_name, where)
    self.execute(sql, params)

  def count(self, table_name, constraints=None, *, extract="index"):
    """Returns the count of records in a table.

    If the default cursor is a tuple or named tuple, this method will work without specifying an
    extract parameter. If it is a dict cursor, it is necessary to specify any value other than
    'index' for extract. This method will not work with cursors that aren't like tuple, namedtuple
    or dict cursors.

    :param table_name: the name of the table to count records on
    :param constraints: can be any construct that can be parsed by SqlWriter.parse_constraints.
    :param extract: the property to pull the count value from the cursor
    :return: the nuber of records matching the constraints
    """
    where, params = self.sql_writer.parse_constraints(constraints)
    sql = "select count(*) as count from {0} where {1};".format(table_name, where or "1 = 1")
    # NOTE: Won't work right with dict cursor
    return self.get_scalar(self.execute(sql, params), 0 if extract == "index" else "count")

  def get_scalar(self, cursor, index=0):
    """Returns a single value from the first returned record from a cursor.

    By default it will get cursor.fecthone()[0] which works with tuples and namedtuples. For dict
    cursor it is necessary to specify index. This method will not work with cursors that aren't
    indexable.

    :param cursor: a cursor object
    :param index: the index of the cursor to return the value from
    """
    if isinstance(index, int):
      return cursor.fetchone()[index]
    else:
      return get_value(cursor.fetchone(), index)


class Upsert(object):
  """This class holds configuration information for performing an upsert."""
  DO_NOTHING = 0x01
  DO_UPDATE = 0x02

  def __init__(self, action=None, target=None, force=False):
    """Initialize the Upsert configuration class.

    :param action: one of Upsert.DO_NOTHING, Upsert.DO_UPDATE. Defaults to DO_NOTHING
    :param target: either a string of an index name, a string of a tuple e.g. '(col1, col2)' or a
                   tuple of column names
    :param force: if False, the upsert will not execute when all of the columns to be updated will
                  not change. If True, no check is done and the update will be performed anyway
    :type force: boolean
    """
    self.action = action or self.DO_NOTHING
    self.target = target
    self.force = force

  @property
  def target_str(self):
    """Returns the string representation of the target property."""
    if isinstance(self.target, tuple):
      return "({})".format(", ".join(self.target))
    else:
      return self.target


class RollbackTransaction(Exception):
  """This Exception class is handled specially by transaction. It will cause the current transaction
  to be rolled back, but the exception won't be reraised.
  """
  pass


@contextmanager
def transaction(data_access):
  """Wrap statements in a transaction. If the statements succeed, commit, otherwise rollback.

  :param data_access: a DataAccess instance
  """
  old_autocommit = data_access.autocommit
  data_access.autocommit = False
  try:
    yield data_access
  except RollbackTransaction as ex:
    data_access.rollback()
  except Exception as ex:
    data_access.rollback()
    raise ex
  else:
    data_access.commit()
  finally:
    data_access.autocommit = old_autocommit


@contextmanager
def autocommit(data_access):
  """Make statements autocommit.

  :param data_access: a DataAccess instance
  """
  if not data_access.autocommit:
    data_access.commit()
  old_autocommit = data_access.autocommit
  data_access.autocommit = True
  try:
    yield data_access
  finally:
    data_access.autocommit = old_autocommit



"""The spreadsheet package contains a standard model for Spreadsheets, and wrappers around several
common formats conforming to the standard model.
"""

from .base import Cell, CellMode
from .csv import CSVWorkbook



"""The spreadsheet.open_document module contains the OpenDocument implementation for the
bidon.spreadsheet model. It requires the ezodf library.
"""
import re

import ezodf

from bidon.spreadsheet.base import CellMode, WorksheetBase, WorkbookBase


__all__ = ["OpenDocumentWorksheet", "OpenDocumentWorkbook"]


_DATE_REGEX = re.compile(r"^(\d{4})-(\d\d)-(\d\d)(?:T(\d\d):(\d\d):(\d\d)(?:\.\d+)?)?$")
_TIME_REGEX = re.compile(r"^PT(\d\d)H(\d\d)M(\d\d(?:\.\d+)?)S$")


class OpenDocumentWorksheet(WorksheetBase):
  """Specialization of WorksheetBase for working with OpenDocument Spreadsheet files."""
  def __init__(self, raw_sheet, ordinal):
    """Initialize the OpenDocumentWorksheet instance."""
    super().__init__(raw_sheet, ordinal)
    self.name = self.raw_sheet.name
    self.nrows = self.raw_sheet.nrows()
    self.ncols = self.raw_sheet.ncols()
    self._raw_rows = None

  def parse_cell(self, cell, coords, cell_mode=CellMode.cooked):
    """Parses a cell according to its cell.value_type."""
    # pylint: disable=too-many-return-statements
    if cell_mode == CellMode.cooked:
      if cell.covered or cell.value_type is None or cell.value is None:
        return None

      vtype = cell.value_type

      if vtype == 'string':
        return cell.value

      if vtype == 'float' or vtype == 'percentage' or vtype == 'currency':
        return cell.value

      if vtype == 'boolean':
        return cell.value

      if vtype == 'date':
        date_tuple = tuple([int(i) if i is not None else 0 \
                            for i in _DATE_REGEX.match(cell.value).groups()])
        return self.tuple_to_datetime(date_tuple)

      if vtype == 'time':
        hour, minute, second = _TIME_REGEX.match(cell.value).groups()
        # TODO: This kills off the microseconds
        date_tuple = (0, 0, 0, int(hour), int(minute), round(float(second)))
        return self.tuple_to_datetime(date_tuple)

      raise ValueError("Unhandled cell type: {0}".format(vtype))
    else:
      return cell

  def get_row(self, row_index):
    """Returns the row at row_index."""
    if self._raw_rows is None:
      self._raw_rows = list(self.raw_sheet.rows())
    return self._raw_rows[row_index]

  def merged_cell_ranges(self):
    """Generates the sequence of merged cell ranges in the format:

    ((col_low, row_low), (col_hi, row_hi))
    """
    for row_number, row in enumerate(self.raw_sheet.rows()):
      for col_number, cell in enumerate(row):
        rspan, cspan = cell.span
        if (rspan, cspan) != (1, 1):
          yield ((col_number, row_number), (col_number + cspan, row_number + rspan))


class OpenDocumentWorkbook(WorkbookBase):
  """Specialization of WorkbookBase for working with OpenDocument Spreadsheet files."""
  def __init__(self, filename):
    """Initializes the OpenDocumentWorkbook instance."""
    super().__init__(filename)
    self.workbook = ezodf.opendoc(self.filename)

  def iterate_sheets(self):
    """Generate the sequence of sheets in the workbook."""
    for sheet in self.workbook.sheets:
      yield sheet

  def get_worksheet(self, raw_sheet, index):
    """Creates an OpenDocumentWorkshet instance from the raw sheet."""
    return OpenDocumentWorksheet(raw_sheet, index)



"""This module contains methods for transforming structured data, such as ElementTree nodes or
dictionaries into other types of values.

It is especially useful for turning JSON or XML data into models.
"""
import bidon.json_patch as JP
from bidon.util import flatten_dict


def get_val(source, extract=None, transform=None):
  """Extract a value from a source, transform and return it."""
  if extract is None:
    raw_value = source
  else:
    raw_value = extract(source)

  if transform is None:
    return raw_value
  else:
    return transform(raw_value)


def get_obj(source, extract=None, child_transform=None, transform=None):
  """Maps an object based on a key->extractor child_transform dict."""
  if extract is None:
    obj = source
  else:
    obj = extract(source)

  if child_transform is None:
    data = obj
  else:
    data = dict()
    for k, v in child_transform.items():
      try:
        data[k] = v(obj)
      except Exception as ex:
        raise Exception("An error occurred with child {0}".format(k)) from ex

  if transform is None:
    return data
  else:
    return transform(data)


def get_lst(source, extract=None, transform=None):
  """Extract a list from a source, transform each item, and return the result."""
  if extract is None:
    raw_list = source
  else:
    raw_list = extract(source)

  if transform is None:
    return raw_list
  else:
    tlist = []
    for idx, item in enumerate(raw_list):
      try:
        tlist.append(transform(item))
      except Exception as ex:
        raise Exception("An error occurred with item #{0}".format(idx)) from ex
    return tlist


def get_composition(source, *fxns):
  """Compose several extractors together, on a source."""
  val = source
  for fxn in fxns:
    val = fxn(val)
  return val


def get_flattened(dct, names, path_joiner="_"):
  """Flatten a child dicts, whose resulting keys are joined by path_joiner.

  E.G. { "valuation": { "currency": "USD", "amount": "100" } } ->
       { "valuation_currency": "USD", "valuation_amount": "100" }
  """
  new_dct = dict()
  for key, val in dct.items():
    if key in names:
      child = {path_joiner.join(k): v for k, v in flatten_dict(val, (key, ))}
      new_dct.update(child)
    else:
      new_dct[key] = dct[key]
  return new_dct


def get_hoisted(dct, child_name):
  """Pulls all of a child's keys up to the parent, with the names unchanged."""
  child = dct[child_name]
  del dct[child_name]
  dct.update(child)
  return dct


def val(extract=None, transform=None):
  """Returns a partial of get_val that only needs a source argument."""
  return lambda source: get_val(source, extract, transform)


def obj(extract=None, child_transform=None, transform=None):
  """Returns a partial of get_obj that only needs a source argument."""
  return lambda source: get_obj(source, extract, child_transform, transform)


def lst(extract=None, transform=None):
  """Returns a partial of get_lst that only needs a source argument."""
  return lambda source: get_lst(source, extract, transform)


def compose(*fxns):
  """Returns a partial of get_composition that only needs a source argument."""
  return lambda source: get_composition(source, *fxns)


def flatten(names, path_joiner="_"):
  """Returns a partial of get_flattened that only needs a dct argument."""
  return lambda dct: get_flattened(dct, names, path_joiner)


def hoisted(child_name):
  """Returns a partial of get_hoisted that only needs a dct argument."""
  return lambda dct: get_hoisted(dct, child_name)


def get_xml_attr(source, name, path=None):
  """Get the XML attribute with name from source. If path is not Mone, it will instead get the
  XML attribute with name from the child indicated by path.
  """
  if path is None:
    return source.attrib[name]
  else:
    return get_xml_attr(get_xml_child(source, path), name)


def get_xml_text(source, path=None):
  """Get the text of the XML node. If path is not None, it will get the text of the descendant of
  source indicated by path.
  """
  if path is None:
    return source.text
  else:
    return get_xml_text(get_xml_child(source, path))


def get_xml_child(source, path):
  """Get the first descendant of source identified by path.

  Path must be either a an xpath string, or a 2-tuple of (xpath, namespace_dict).
  """
  if isinstance(path, (tuple, list)):
    return source.find(*path)
  else:
    return source.find(path)


def get_xml_children(source, path):
  """Get all the descendants of source identified by path.

  Path must be either a an xpath string, or a 2-tuple of (xpath, namespace_dict).
  """
  if isinstance(path, (tuple, list)):
    return source.findall(*path)
  else:
    return source.findall(path)


def xml_attr(name, path=None):
  """Returns a partial of get_xml_attr that only needs a source argument."""
  return lambda source: get_xml_attr(source, name, path)


def xml_text(path=None):
  """Returns a partial of get_xml_text that only needs a source argument."""
  return lambda source: get_xml_text(source, path)


def xml_child(path):
  """Returns a partial of get_xml_child that only needs a source argument."""
  return lambda source: get_xml_child(source, path)


def xml_children(path):
  """Returns a partial of get_xml_children that only needs a source argument."""
  return lambda source: get_xml_children(source, path)


def get_json_val(source, path, *, ignore_bad_path=False):
  """Get the nested value identified by the json path, rooted at source."""
  try:
    return JP.find(source, path)
  except JP.JSONPathError as ex:
    if ignore_bad_path:
      return None
    else:
      raise


def get_json_vals(source, path):
  """Get all the nested values identified by the json path, rooted at source."""
  yield from JP.find_all(source, path)


def json_val(path, *, ignore_bad_path=False):
  """Returns a partial of get_json_val that only needs a source argument."""
  return lambda source: get_json_val(source, path, ignore_bad_path=ignore_bad_path)


def json_vals(path):
  """Returns a partial of get_json_vals that only needs a source argument."""
  return lambda source: get_json_vals(source, path)



"""Contains the FieldMapping class."""
from collections import namedtuple

from bidon.util import has_value
from bidon.util.convert import identity


class FieldMapping(object):
  """This class contains a definition for mapping a field in a structure to a new structure,
  potentially changing its name and value.
  """
  def __init__(self, source_name, destination_name=None, convert=None, is_required=True):
    """Initialize the FieldMapping instance."""
    self.source_name = source_name
    self.destination_name = destination_name or source_name
    self.convert = convert or identity
    self.is_required = is_required

  def get_value(self, source):
    """Apply self.convert to the source. The parameter passed to convert depends on
    self.source_name. If source_name is given, self.convert(getattr(source, source_name)) is called,
    otherwise self.convert(source) is called.
    """
    if self.source_name is None:
      present, value = True, self.convert(source)
      converted = True
    else:
      present, value = has_value(source, self.source_name)
      converted = False

    if not present or value is None:
      if self.is_required:
        raise ValueError("required value not present")
      else:
        return None
    else:
      if converted:
        return value
      else:
        return self.convert(value)

  @classmethod
  def get_namedtuple(cls, field_mappings, name="Record"):
    """Gets a namedtuple class that matches the destination_names in the list of field_mappings."""
    return namedtuple(name, [fm.destination_name for fm in field_mappings])

  @classmethod
  def get_namedtuple_factory(cls, field_mappings, name="Record"):
    """Gets a method that will convert a dictionary to a namedtuple, as defined by
    get_namedtuple(field_mappings).
    """
    ntup = cls.get_namedtuple(field_mappings, name)
    return lambda data: ntup(**data)

  @classmethod
  def transfer(cls, field_mappings, source, destination_factory):
    """Convert a record to a dictionary via field_mappings, and pass that to destination_factory."""
    data = dict()
    for index, field_mapping in enumerate(field_mappings):
      try:
        data[field_mapping.destination_name] = field_mapping.get_value(source)
      except Exception as ex:
        raise Exception(
          "Error with mapping #{0} '{1}'->'{2}': {3}".format(
            index,
            field_mapping.source_name,
            field_mapping.destination_name, ex)) from ex
    return destination_factory(data)

  @classmethod
  def transfer_all(cls, field_mappings, sources, destination_factory=None):
    """Calls cls.transfer on all records in sources."""
    for index, source in enumerate(sources):
      try:
        yield cls.transfer(field_mappings, source, destination_factory or (lambda x: x))
      except Exception as ex:
        raise Exception("Error with source #{0}: {1}".format(index, ex)) from ex



"""Contains ModelAccess class and helper methods."""
from bidon.util.date import local_now, utc_now
from .data_access import DataAccess, Upsert


class ModelAccess(DataAccess):
  """A model-based specialization of DataAccess

  Provides methods for CRUDing objects derived from bidon.db.model.ModelBase.
  """
  def _find_model(self, constructor, table_name, constraints=None, *, columns=None, order_by=None):
    """Calls DataAccess.find and passes the results to the given constructor."""
    data = self.find(table_name, constraints, columns=columns, order_by=order_by)
    return constructor(data) if data else None

  def _find_models(self, constructor, table_name, constraints=None, *, columns=None,
                   order_by=None, limiting=None):
    """Calls DataAccess.find_all and passes the results to the given constructor."""
    for record in self.find_all(table_name, constraints, columns=columns, order_by=order_by,
                                limiting=limiting):
      yield constructor(record)

  def find_model(self, constructor, constraints=None, *, columns=None, table_name=None,
                 order_by=None):
    """Specialization of DataAccess.find that returns a model instead of cursor object."""
    return self._find_model(constructor, table_name or constructor.table_name, constraints,
                            columns=columns, order_by=order_by)

  def find_models(self, constructor, constraints=None, *, columns=None, order_by=None,
                  limiting=None, table_name=None):
    """Specialization of DataAccess.find_all that returns models instead of cursor objects."""
    return self._find_models(
      constructor, table_name or constructor.table_name, constraints, columns=columns,
      order_by=order_by, limiting=limiting)

  def page_models(self, constructor, paging, constraints=None, *, columns=None, order_by=None):
    """Specialization of DataAccess.page that returns models instead of cursor objects."""
    records, count = self.page(constructor.table_name, paging, constraints, columns=columns,
                               order_by=order_by)
    return ([constructor(r) for r in records], count)

  def find_model_by_id(self, constructor, id_, *, columns=None):
    """Searches for a model by id, according to its class' primary_key_name.

    If primary_key_name is a tuple, id_ must be a tuple with a matching length.
    """
    return self.find_model(
      constructor, get_id_constraints(constructor.primary_key_name, id_), columns=columns)

  def refresh_model(self, model, *, overwrite=False):
    """Pulls the model's record from the database. If overwrite is True, the model values are
    overwritten and returns the model, otherwise a new model instance with the newer record is
    returned.
    """
    new_model = self.find_model_by_id(model.__class__, model.primary_key)
    if overwrite:
      model.update(new_model.to_dict(use_default_excludes=False))
      return model
    else:
      return new_model

  def update_model(self, model, *, include_keys=None):
    """Updates a model.

    :include_keys: if given, only updates the given attributes. Otherwise, updates all non-id
    attributes.
    """
    id_constraints = get_model_id_constraints(model)

    if include_keys is None:
      include_keys = set(
        model.attrs.keys()).difference(model.exclude_keys_sql).difference(id_constraints.keys())

    # If include_keys was not null but was empty
    if not include_keys:
      return model

    values = model.to_dict(include_keys=include_keys)
    returnings = []

    _, updated_ts = model.timestamps if model.timestamps else (None, None)

    if updated_ts and updated_ts not in values:
      values[updated_ts] = utc_now() if self.core.supports_timezones else local_now()
      returnings.append(updated_ts)

    returning = ", ".join(returnings)
    cr = self.update(model.table_name, values, id_constraints, returning=returning)

    if returning and self.core.supports_returning_syntax:
      rec = cr.fetchone()
      for idx, attr_name in enumerate(returnings):
        setattr(model, attr_name, rec[idx])

    return model

  def insert_model(self, model, *, upsert=None):
    """Inserts a record for the given model.

    If model's primary key is auto, the primary key will be set appropriately.
    """
    pkname = model.primary_key_name
    include_keys = set(model.attrs.keys()).difference(model.exclude_keys_sql)

    if model.primary_key_is_auto:
      if pkname in include_keys:
        include_keys.remove(pkname)
    else:
      if isinstance(pkname, str):
        include_keys.add(pkname)
      else:
        include_keys.update(set(pkname))

    data = model.to_dict(include_keys=include_keys)
    returnings = []

    if model.primary_key_is_auto:
      returnings.append(pkname)
    if model.timestamps:
      returnings.extend(ts_name for ts_name in model.timestamps if ts_name)

    returning = ", ".join(returnings)
    cr = self.insert(model.table_name, data, returning=returning, upsert=upsert)

    if self.core.supports_returning_syntax:
      if returning:
        rec = cr.fetchone()
        if rec:
          for idx, attr_name in enumerate(returnings):
            setattr(model, attr_name, rec[idx])
    else:
      if model.primary_key_is_auto:
        setattr(model, model.primary_key_name, cr.lastrowid)

    return model

  def delete_model(self, model_or_type, id_=None):
    """Deletes a model.

    :model_or_type: if a model, delete that model. If it is a ModelBase subclass, id_ must be
    specified, and the associated record is deleted.
    """
    if not id_:
      constraints = get_model_id_constraints(model_or_type)
    else:
      constraints = get_id_constraints(model_or_type.primary_key_name, id_)
    self.delete(model_or_type.table_name, constraints)

    return model_or_type

  def find_or_build(self, constructor, props):
    """Looks for a model that matches the given dictionary constraints. If it is not found, a new
    model of the given type is constructed and returned.
    """
    model = self.find_model(constructor, props)
    return model or constructor(**props)

  def find_or_create(self, constructor, props, *, comp=None):
    """Looks for a model taht matches the given dictionary constraints. If it is not found, a new
    model of the given type is created and saved to the database, then returned.
    """
    model = self.find_model(constructor, comp or props)
    if model is None:
      model = constructor(**props)
      self.insert_model(model)
    return model

  def find_or_upsert(self, constructor, props, *, comp=None, return_status=False):
    """This finds or upserts a model with an auto primary key, and is a bit more flexible than
    find_or_create.

    First it looks for the model matching either comp, or props if comp is None.

    If not found, it will try to upsert the model, doing nothing. If the returned model is new,
    meaning it's primary key is not set, then the upsert was unable to create the model, meaning
    there was a conflict. If there is a conflict, find model is run again, and this time it
    will succeed*. Otherwise, the constructed model is returned.

    *this is not entirely true. It's possible that the upsert returns with None, meaning that a
    record was created between the first find and the upsert, and then deleted between the upsert
    and the second find. This situation is out of the scope of this method. A possible solution
    would be to repeat the find/uspert cycle until a model can be returned, but I'm going to avoid
    that for simplicty for now.

    :param constructor: the model constructor
    :param props: the properties to construct the model with if not found
    :param comp: the properties to search for the model with. If None, props is used
    :param return_status: if True, a 2-tuple of (model, status) is returned, where status is what
                          occurred with the model. Either 'found', 'created' or 'duplicate'.
    """
    model = self.find_model(constructor, comp or props)
    status = _UPSERT_STATUS_FOUND

    if model is None:
      model = constructor(**props)
      status = _UPSERT_STATUS_CREATED
      self.insert_model(model, upsert=Upsert(Upsert.DO_NOTHING))
      if model.is_new:
        model = self.find_model(constructor, comp or props)
        status = _UPSERT_STATUS_DUPLICATE

    if return_status:
      return (model, status)
    else:
      return model


def get_model_id_constraints(model):
  """Returns constraints to target a specific model."""
  pkname = model.primary_key_name
  pkey = model.primary_key
  return get_id_constraints(pkname, pkey)


def get_id_constraints(pkname, pkey):
  """Returns primary key consraints.

  :pkname: if a string, returns a dict with pkname=pkey. pkname and pkey must be enumerables of
  matching length.
  """
  if isinstance(pkname, str):
    return {pkname: pkey}
  else:
    return dict(zip(pkname, pkey))


_UPSERT_STATUS_FOUND = "found"
_UPSERT_STATUS_CREATED = "created"
_UPSERT_STATUS_DUPLICATE = "duplicate"



"""Contains the StreamWriter class."""
from collections import OrderedDict
from contextlib import contextmanager
from xml.sax.saxutils import XMLGenerator


def _normalize_name(name):
  """Returns a normalized element tag or attribute name. A tag name or attribute name can be given
  either as a string, or as a 2-tuple. If a 2-tuple, it is interpreted as (namespace, name) and is
  returned joined by a colon ":"

  :name: a string or a 2-tuple of strings
  """
  if isinstance(name, tuple):
    ns, nm = name
    return "{}:{}".format(ns, nm)
  else:
    return name


def _normalize_value(val, null):
  """Returns a normalized attribute value. A value can be either a string, or None. If None, the
  value of :null: will be returned instead.

  :val: a string or None
  :null: a string to be returned when :val: is None
  """
  if val is None:
    return null
  else:
    return str(val)


def _normalize_attrs(attrs, null=None):
  """Returns an OrderedDict of attirbutes with the keys normalized by _normalize_name and the values
  normalized by _normalize_value. The value of :null: is passed to _normalize_value. However, if
  :null: is None, any values that are None are left out of the returned dict.

  An OrderedDict is returned so that the attributes are always printed in the same order for similar
  objects.

  :attrs: a dict whose keys can be passed to _normalize_name and whose values can be passed to
          _normalize_value
  :null: either a string to be put in place of values that are None, or None, which will cause
         values that are None to be left out of the returned dict
  """
  if attrs:
    items = sorted((_normalize_name(k), _normalize_value(v, null)) for (k, v) in attrs.items())

    if null is None:
      return OrderedDict((k, v) for (k, v) in items if v is not None)
    else:
      return OrderedDict(items)
  else:
    return {}


class StreamWriter(object):
  """This class provides methods for writing out an XML tree in a stream, rather than building it
  up in memory and then dumping it. The API is low-level, but it allows for very low memory usage,
  which is necessary when generating large files.

  This is implemented as a thin wrapper around xml.sax.saxutils.XMLGenerator.
  """
  def __init__(self, out, encoding="UTF-8", indent=None):
    """Initializes the StreamWriter instance.

    :out: a file-like object that will be written to
    :encodig: the encoding for output. Defaults to UTF-8
    :indent: a whitespace string used for indenting new lines. If None (the default) then no
             extra whitespace will be added to the document.
    """
    self._writer = XMLGenerator(out, encoding)
    self._indent = indent
    self._open_elements = []

  def __enter__(self):
    """Entry point for use as a context manager."""
    self.start()
    return self

  def __exit__(self, ex_type, ex_val, ex_tb):
    """Exit point for use as a context manager."""
    if ex_type is None:
      self.end()

  def start(self):
    """Calls startDocument on the underlying XMLGenerator."""
    self._writer.startDocument()

  def end(self):
    """Closes any remaining open elements and calls endDocument on the underlying XMLGenerator."""
    while self._open_elements:
      self.close()

    self._writer.endDocument()

  def open(self, name, attrs=None, *, close=False):
    """Writes an opening element.

    :name: the name of the element
    :attrs: a dict of attributes
    :close: if True, close will be called immediately after writing the element
    """
    self._pad()
    self._writer.startElement(_normalize_name(name), _normalize_attrs(attrs))
    self._newline()
    self._open_elements.append(name)

    if close:
      self.close()

  def close(self, name=None):
    """Closes the most recently opened element.

    :name: if given, this value must match the name given for the most recently opened element. This
           is primarily here for providing quick error checking for applications
    """
    tag = self._open_elements.pop()
    if name is not None and name != tag:
      raise Exception("Tag closing mismatch")
    self._pad()
    self._writer.endElement(_normalize_name(tag))
    self._newline()

  def characters(self, characters):
    """Writes content for a tag.

    :characters: the characters to write
    """
    self._pad()
    self._writer.characters(str(characters))
    self._newline()

  def whitespace(self, whitespace):
    """Writes ignorable whitespace.

    :whitespace: the whitespace to write
    """
    self._writer.ignorableWhitespace(whitespace)

  @contextmanager
  def element(self, name, attrs=None):
    """This method is a context manager for writing and closing an element."""
    self.open(name, attrs)
    yield
    self.close()

  @contextmanager
  def no_inner_space(self, *, outer=True):
    """Default spacing for all things written is ignored in this context.

    :outer: boolean, if True the typical padding and newline are added before the first and after
            the last things written
    """
    if outer:
      self._pad()

    indent_was = self._indent
    self._indent = None

    try:
      yield
    finally:
      self._indent = indent_was

    if outer:
      self._newline()

  def content(self, name, attrs=None, characters=None):
    """Writes an element, some content for the element, and then closes the element, all without
    indentation.

    :name: the name of the element
    :attrs: a dict of attributes
    :characters: the characters to write
    """
    with self.no_inner_space(outer=True):
      with self.element(name, attrs):
        if characters:
          self.characters(characters)

  def _pad(self):
    """Pads the output with an amount of indentation appropriate for the number of open element.

    This method does nothing if the indent value passed to the constructor is falsy.
    """
    if self._indent:
      self.whitespace(self._indent * len(self._open_elements))

  def _newline(self):
    """Writes a newline to output.

    This method does nothing if the indent value passed to the constructor is falsy.
    """
    if self._indent:
      self.whitespace("\n")



"""This module contains classes for interacting with the database at a high-level."""

from .data_access import DataAccess, RollbackTransaction, Upsert, autocommit, transaction
from .model_access import ModelAccess
from . import pg_advisory_lock



"""The data_access_core module contains the DataAccessCoreBase class and some useful derivations,
as well as methods for constructing cores for common database systems, including PostgreSQL, SQLite,
and MySQL.
"""
from .sql_writer import SqlWriter


__all__ = ["DataAccessCoreBase",
           "InjectedDataAccessCore",
           "get_pg_core",
           "get_pooled_pg_core",
           "get_sqlite_core",
           "get_mysql_core"]


SQL_CAST = "cast({0} as {1})"


class DataAccessCoreBase(object):
  """Contains settings for DataAccess that allows different database engines to be supported."""
  def __init__(self, sql_writer=("%({0})s", "%s", SQL_CAST), empty_params=None,
               supports_timezones=True, supports_returning_syntax=True):
    """Initialize a DataAccessCoreBase instance.

    :param sql_writer: either a SqlWriter instance or a tuple of constructor arguments.
    :param empty_params: the value that should be used when the params passed to exec are None
    :param supports_timezones: whether or not the database supports timezones in timestamps
    :type supports_timezones: boolean
    :param supports_returning_syntax: whether or not the database supports the returning syntax
    :type supports_returning_syntax: boolean
    """
    self.sql_writer = sql_writer if isinstance(sql_writer, SqlWriter) else SqlWriter(*sql_writer)
    self.empty_params = empty_params
    self.supports_timezones = supports_timezones
    self.supports_returning_syntax = supports_returning_syntax

  def open(self):
    """Open a connection.

    This is an abstract method.
    """
    raise NotImplementedError()

  def close(self, connection, *, commit=True):
    """Close a connection, commiting pending changes if commit is True.

    This is an abstract method.
    """
    raise NotImplementedError()

  def get_autocommit(self, connection):
    """Gets connection's autocommit value.

    This is an abstract method.
    """
    raise NotImplementedError()

  def set_autocommit(self, connection, value):
    """Sets connection's autocommit value.

    This is an abstract method.

    :param connection: a connection instance
    :param value: the value to set autocommit to
    """
    raise NotImplementedError()


class InjectedDataAccessCore(DataAccessCoreBase):
  """An implementation of DataAccessCoreBase where the open and close methods are passed to the
  constructor.
  """
  def __init__(self, opener, closer, sql_writer=("%({0})s", "%s", SQL_CAST), *, empty_params=None,
               supports_timezones=True, supports_returning_syntax=True, get_autocommit=None,
               set_autocommit=None):
    super().__init__(sql_writer, empty_params, supports_timezones, supports_returning_syntax)
    self.opener = opener
    self.closer = closer
    self._get_autocommit = get_autocommit
    self._set_autocommit = set_autocommit

  def open(self):
    """Open and return a connection using the opener method passed to the constructor."""
    return self.opener()

  def close(self, connection, *, commit=True):
    """Close the connection using the closer method passed to the constructor."""
    if commit:
      connection.commit()
    else:
      connection.rollback()
    self.closer(connection)

  def get_autocommit(self, connection):
    """Gets connection's autocommit value."""
    return self._get_autocommit(connection)

  def set_autocommit(self, connection, value):
    """Sets connection's autocommit value."""
    return self._set_autocommit(connection, value)


def default_connection_closer(connection):
  """The default connection closer complying with DB API 2.0 standards."""
  connection.close()


def get_pg_autocommit(cn):
  """PostgreSQL autocommit getter for core."""
  return cn.autocommit


def set_pg_autocommit(cn, autocommit):
  """PostgreSQL autocommit setter for core."""
  cn.autocommit = autocommit


def get_pg_core(connection_string, *, cursor_factory=None, edit_connection=None):
  """Creates a simple PostgreSQL core. Requires the psycopg2 library."""
  import psycopg2 as pq
  from psycopg2.extras import NamedTupleCursor

  def opener():
    """Opens a single PostgreSQL connection with the scope-captured connection string."""
    cn = pq.connect(connection_string)
    cn.cursor_factory = cursor_factory or NamedTupleCursor
    if edit_connection:
      edit_connection(cn)
    return cn

  return InjectedDataAccessCore(
    opener,
    default_connection_closer,
    ("%({0})s", "%s", "{0}::{1}"),
    empty_params=None,
    supports_timezones=True,
    supports_returning_syntax=True,
    get_autocommit=get_pg_autocommit,
    set_autocommit=set_pg_autocommit)


def get_pooled_pg_core(connection_string, pool_size=None, *, cursor_factory=None,
                       edit_connection=None, threaded=True):
  """Creates a pooled PostgreSQL core. Requires the psycopg2 library.

  :pool_size: must be a 2-tuple in the form (min_connections, max_connections).
  """
  from psycopg2.extras import NamedTupleCursor
  from psycopg2.pool import ThreadedConnectionPool as TPool, SimpleConnectionPool as SPool
  if not pool_size:
    pool_size = (5, 10)
  if threaded:
    pool = TPool(pool_size[0], pool_size[1], connection_string)
  else:
    pool = SPool(pool_size[0], pool_size[1], connection_string)

  def opener():
    """Gets a PostgreSQL connection from the scope-captured connection pool."""
    cn = pool.getconn()
    cn.cursor_factory = cursor_factory or NamedTupleCursor
    if edit_connection:
      edit_connection(cn)
    return cn

  def closer(connection):
    """Returns a connection to the scope-captured connection pool."""
    pool.putconn(connection)

  return InjectedDataAccessCore(
    opener,
    closer,
    ("%({0})s", "%s", "{0}::{1}"),
    empty_params=None,
    supports_timezones=True,
    supports_returning_syntax=True,
    get_autocommit=get_pg_autocommit,
    set_autocommit=set_pg_autocommit)


def get_sqlite_autocommit(cn):
  """SQLite autocommit getter for core."""
  return cn.isolation_level is None


def set_sqlite_autocommit(cn, autocommit):
  """SQLite autocommit setter for core."""
  if isinstance(autocommit, bool):
    cn.isolation_level = None if autocommit else ""
  else:
    cn.isolation_level = autocommit


def get_sqlite_core(connection_string, *, cursor_factory=None, edit_connection=None):
  """Creates a simple SQLite3 core."""
  import sqlite3 as sqlite

  def opener():
    """Opens a single connection with the scope-captured connection string."""
    cn = sqlite.connect(connection_string)
    if cursor_factory:
      cn.row_factory = cursor_factory
    if edit_connection:
      edit_connection(cn)
    return cn

  return InjectedDataAccessCore(
    opener,
    default_connection_closer, (":{0}", "?", SQL_CAST),
    empty_params=[],
    supports_timezones=True,
    supports_returning_syntax=False,
    get_autocommit=get_sqlite_autocommit,
    set_autocommit=set_sqlite_autocommit)


def get_mysql_autocommit(cn):
  """MySQL autocommit getter for core."""
  return cn.get_autocommit()


def set_mysql_autocommit(cn, autocommit):
  """MySQL autocommit setter for core."""
  cn.autocommit(autocommit)


def get_mysql_core(connection_args, *, cursor_factory=None, edit_connection=None):
  """Creates a simple MySQL core. Requires the pymysql library."""
  import pymysql

  def opener():
    """Opens a single connection with the scope-captured connection string."""
    cn = pymysql.connect(**connection_args)
    if cursor_factory:
      cn.cursorclass = cursor_factory
    if edit_connection:
      edit_connection(cn)
    return cn

  return InjectedDataAccessCore(
    opener,
    default_connection_closer,
    ("%({0})s", "%s", SQL_CAST),
    empty_params=None,
    supports_timezones=False,
    supports_returning_syntax=False,
    get_autocommit=get_mysql_autocommit,
    set_autocommit=set_mysql_autocommit)



import unittest

from bidon import json_patch as jp


class JSONPatchTestCase(unittest.TestCase):
  def setUp(self):
    self.document = {
      "name": "Trey",
      "computers": ["13\" macbook", "15\" macbook pro", "iPad 1"],
      "age": 31,
      "spouse": {
        "name": "Julie",
        "age": 29,
        "computers": ["13\" macbook pro", "iPhone 6"],
      },
      "children": ["Asher"],
      "path/test": {
        "goes~to": "o7911"
      }
    }

  def apply_patch(self, patch):
    jp.apply_patch(self.document, patch)
    return self.document

  def test_path_escapes(self):
    r, i = jp.resolve_path(self.document, "/path~1test/goes~0to")
    self.assertEqual(r[i], "o7911")

  def test_add(self):
    patch = jp.Patch(op="add", path="/eye_color", value="blue")
    self.assertEqual(self.apply_patch(patch)["eye_color"], "blue")

    patch = jp.Patch(op="add", path="/spouse/eye_color", value="brown")
    self.assertEqual(self.apply_patch(patch)["spouse"]["eye_color"], "brown")

    patch = jp.Patch(op="add", path="/computers/2", value="iPhone 5")
    self.assertEqual(self.apply_patch(patch)["computers"][2], "iPhone 5")

    patch = jp.Patch(op="add", path="/computers/", value="iPhone 6")
    self.assertEqual(self.apply_patch(patch)["computers"][-1], "iPhone 6")

    patch = jp.Patch(op="add", path="/computers/~", value="iPhone 6 Plus")
    self.assertEqual(self.apply_patch(patch)["computers"][-1], "iPhone 6 Plus")

    with self.assertRaises(Exception):
      self.apply_patch(jp.Patch(op="add", path="/name", value=None))

  def test_remove(self):
    patch = jp.Patch(op="remove", path="/age")
    self.assertTrue("age" not in self.apply_patch(patch))

    patch = jp.Patch(op="remove", path="/computers/0")
    pv = self.document["computers"][0]
    self.assertNotEqual(self.apply_patch(patch)["computers"][0], pv)

    with self.assertRaises(Exception):
      self.apply_patch(jp.Patch(op="remove", path="eye_color"))

  def test_replace(self):
    patch = jp.Patch(op="replace", path="/age", value=32)
    self.assertEqual(self.apply_patch(patch)["age"], 32)

    with self.assertRaises(Exception):
      self.apply_patch(jp.Patch(op="replace", path="/eye_color", value="green"))

  def test_merge(self):
    patch = jp.Patch(op="merge", path="/spouse", value=dict(age=30, name="Juliana", hair_length="long"))
    self.apply_patch(patch)
    s = self.document["spouse"]
    self.assertEqual(s["age"], 30)
    self.assertEqual(s["name"], "Juliana")
    self.assertEqual(s["hair_length"], "long")

  def test_copy(self):
    patch = jp.Patch(op="copy", src="/computers", path="/spouse/computers")
    self.apply_patch(patch)
    self.assertEqual(self.document["computers"], self.document["spouse"]["computers"])

  def test_move(self):
    patch = jp.Patch(op="move", src="/children", path="/spouse/children")
    mv = self.document["children"]
    self.assertEqual(self.apply_patch(patch)["spouse"]["children"], mv)

  def test_test(self):
    d = self.document
    self.assertTrue(jp.apply_patch(d, jp.Patch(op="test", path="/age")))
    self.assertTrue(jp.apply_patch(d, jp.Patch(op="test", path="/age", value=31)))
    self.assertFalse(jp.apply_patch(d, jp.Patch(op="test", path="/age", value=32)))
    self.assertFalse(jp.apply_patch(d, jp.Patch(op="test", path="/hamburger_count")))
    self.assertFalse(jp.apply_patch(d, jp.Patch(op="test", path="/hamburger_count", value="the, all of")))

  def test_set_remove(self):
    v = "13\" macbook pro"
    patch = jp.Patch(op="setremove", path="/spouse/computers", value=v)
    self.assertTrue(v in self.document["spouse"]["computers"])
    self.assertFalse(v in self.apply_patch(patch)["spouse"]["computers"])

  def test_set_add(self):
    v = "iPad Air"
    patch = jp.Patch(op="setadd", path="/computers", value=v)
    self.assertFalse(v in self.document["computers"])
    self.assertTrue(self.apply_patch(patch)["computers"].count(v) == 1)
    # Adding the same item multiple times to a set has no effect
    self.assertTrue(self.apply_patch(patch)["computers"].count(v) == 1)

  def test_find(self):
    self.assertEqual(jp.find(self.document, "/spouse/name"), "Julie")
    self.assertEqual(jp.find(self.document, "/spouse/computers/1"), "iPhone 6")

  def test_find_all(self):
    doc = dict(
      person=dict(
        name="Trey",
        things=[
          dict(id=1, name="One"),
          dict(id=2, name="Two")
        ]))
    self.assertEqual(set(jp.find_all(doc, "/person/things/*/name")), {"One", "Two"})



import unittest
from datetime import date, time, datetime
from numbers import Number

from bidon.spreadsheet import Cell


class SpreadsheetTestCase(unittest.TestCase):
  def check_workbook(self, wb_class, open_args, ncols, nrows, date_type, time_type, datetime_type, has_merges, has_notes):
    wb = wb_class(*open_args)
    self.assertEqual(1, len(wb.sheets()))
    sh = wb.sheets(0)
    self.assertEqual(ncols, sh.ncols)
    self.assertEqual(nrows, sh.nrows)
    self.assertEqual(sh.get_cell((0, 0)), "Name")
    self.assertIsInstance(sh.get_cell((1, 1)), date_type)
    self.assertIsInstance(sh.get_cell((2, 1)), time_type)
    self.assertIsInstance(sh.get_cell((3, 1)), Number)
    self.assertIsInstance(sh.get_cell((4, 1)), datetime_type)
    if has_merges:
      merged_ranges = list(sh.merged_cell_ranges())
      self.assertEqual(3, len(merged_ranges))
      self.assertTrue(((0, 2), (2, 3)) in merged_ranges)
      self.assertTrue(((2, 2), (3, 4)) in merged_ranges)
      self.assertTrue(((3, 2), (5, 4)) in merged_ranges)
    if has_notes:
      self.assertIsNone(sh.get_note((0, 0)))
      self.assertIsNotNone(sh.get_note((0, 3)))
    return wb

  def test_xlsx(self):
    from bidon.spreadsheet.excel import ExcelWorkbook as WB
    wb = self.check_workbook(WB, ["tests/fixtures/test.xlsx", False], 5, 4, date, time, datetime, True, True)
    ct = wb.sheets(0).to_cell_table()
    self.assertEqual(4, len(ct))
    self.assertEqual(5, len(ct[0]))
    self.assertIsInstance(ct[0][0], Cell)
    self.assertEqual(ct[0][0].value, "Name")
    self.assertEqual(ct[3][0].note.strip(), "A comment my friends")

  def test_xls(self):
    from bidon.spreadsheet.excel import ExcelWorkbook as WB
    self.check_workbook(WB, ["tests/fixtures/test.xls", True], 5, 4, date, time, datetime, True, True)

  def test_ods(self):
    from bidon.spreadsheet.open_document import OpenDocumentWorkbook as WB
    self.check_workbook(WB, ["tests/fixtures/test.ods"], 6, 6, date, time, datetime, True, False)

  def test_csv(self):
    from bidon.spreadsheet.csv import CSVWorkbook as WB
    self.check_workbook(WB, ["tests/fixtures/test.csv"], 5, 4, str, str, str, False, False)

    import gzip
    with gzip.open("tests/fixtures/test.csv.gz", "rt") as rf:
      self.check_workbook(WB, [rf], 5, 4, str, str, str, False, False)

    self.check_workbook(WB, ["tests/fixtures/test.csv.gz"], 5, 4, str, str, str, False, False)



"""The spreadsheet.excel module contains the Excel implementation for the bidon.spreadsheet model
for working with .xls and .xlsx documents. It requires the xlrd library.
"""
import xlrd

from bidon.spreadsheet.base import CellMode, WorksheetBase, WorkbookBase


__all__ = ["ExcelWorksheet", "ExcelWorkbook"]


class ExcelWorksheet(WorksheetBase):
  """Specialization of WorksheetBase for working with excel documents."""
  def __init__(self, raw_sheet, ordinal, handle_ambiguous_date):
    """Initialize the ExcelWorksheet instance."""
    super().__init__(raw_sheet, ordinal)
    self.name = self.raw_sheet.name
    self.nrows = self.raw_sheet.nrows
    self.ncols = self.raw_sheet.ncols
    self.handle_ambiguous_date = handle_ambiguous_date

  def parse_cell(self, cell, coords, cell_mode=CellMode.cooked):
    """Parses a cell according to the cell.ctype."""
    # pylint: disable=too-many-return-statements
    if cell_mode == CellMode.cooked:
      if cell.ctype == xlrd.XL_CELL_BLANK:
        return None

      if cell.ctype == xlrd.XL_CELL_BOOLEAN:
        return cell.value

      if cell.ctype == xlrd.XL_CELL_DATE:
        if self.handle_ambiguous_date:
          try:
            return self._parse_date(cell.value)
          except xlrd.xldate.XLDateAmbiguous:
            return self.handle_ambiguous_date(cell.value)
        else:
          return self._parse_date(cell.value)

      if cell.ctype == xlrd.XL_CELL_EMPTY:
        return None

      if cell.ctype == xlrd.XL_CELL_ERROR:
        return cell.value

      if cell.ctype == xlrd.XL_CELL_NUMBER:
        return cell.value

      if cell.ctype == xlrd.XL_CELL_TEXT:
        return cell.value

      raise ValueError("Unhandled cell type {0}".format(cell.ctype))
    else:
      return cell

  def get_row(self, row_index):
    """Returns the row at row_index."""
    return self.raw_sheet.row(row_index)

  def merged_cell_ranges(self):
    """Generates the sequence of merged cell ranges in the format:

    ((col_low, row_low), (col_hi, row_hi))
    """
    for rlo, rhi, clo, chi in self.raw_sheet.merged_cells:
      yield ((clo, rlo), (chi, rhi))

  def get_note(self, coords):
    """Get the note for the cell at the given coordinates.

    coords is a tuple of (col, row)
    """
    col, row = coords
    note = self.raw_sheet.cell_note_map.get((row, col))
    return note.text if note else None

  def _parse_date(self, cell_value):
    """Attempts to parse a cell_value as a date."""
    date_tuple = xlrd.xldate_as_tuple(cell_value, self.raw_sheet.book.datemode)
    return self.tuple_to_datetime(date_tuple)


class ExcelWorkbook(WorkbookBase):
  """Specialization of WorkbookBase for working with Excel workbooks."""
  def __init__(self, filename, formatting_info=False, handle_ambiguous_date=None):
    """Initialize the ExcelWorkbook instance."""
    super().__init__(filename)
    self.workbook = xlrd.open_workbook(self.filename, formatting_info=formatting_info)
    self.handle_ambiguous_date = handle_ambiguous_date

  def iterate_sheets(self):
    """Generates the sequence of sheets in the workbook."""
    for sheet in self.workbook.sheets():
      yield sheet

  def get_worksheet(self, raw_sheet, index):
    """Creates an ExcelWorksheet instance from a raw sheet."""
    return ExcelWorksheet(raw_sheet, index, self.handle_ambiguous_date)



"""The validation module contains the Validation and Validator classes."""
import bidon.util as util


__all__ = ["Validation", "Validator"]


class Validation(object):
  """A simple validation mechanism, designed for use by bidon.db.model.ModelBase."""
  def __init__(self, callback, property_name=None, message=None, *, is_simple=True,
               is_property_specific=True):
    """
    Initializes the Validation instance.

    :is_simple: when True the validation requires no outside information to process, if False,
    outside information (such as a database connection) is required.
    :is_property_specific: when True, the value of the property under question is passed to the
    validation callback, when false the entire model is passed.
    """
    self.property_name = property_name
    self.callback = callback
    self.message = message
    self.is_simple = is_simple
    self.filters = []
    self.pretty_property_name = None
    self.is_property_specific = is_property_specific

  def add_filter(self, filter_):
    """Filters are methods that accept a model and return True if the model should be validated or
    False if it should not. This makes it possible to exempt models from validations when
    necessary.
    """
    self.filters.append(filter_)
    return self

  def set_pretty_property_name(self, ppn):
    """Set the pretty property name for the validation and return the validation instance."""
    self.pretty_property_name = ppn
    return self

  def is_valid(self, model, validator=None):
    """Returns true if the model passes the validation, and false if not. Validator must be
    present_optional if validation is not 'simple'.
    """
    if self.property_name and self.is_property_specific:
      arg0 = getattr(model, self.property_name)
    else:
      arg0 = model

    if self.is_simple:
      is_valid = self.callback(arg0)
    else:
      is_valid = self.callback(arg0, validator)

    return (is_valid, None if is_valid else (self.message or "is invalid"))

  def validate(self, model, validator=None):
    """Checks the model against all filters, and if it shoud be validated, runs the validation. if
    the model is invalid, an error is added to the model. Then the validity value is returned.
    """
    for filter_ in self.filters:
      if not filter_(model):
        return True

    is_valid, message = self.is_valid(model, validator)
    if not is_valid:
      model.add_error(self.pretty_property_name or self.property_name, message)
    return is_valid

  @staticmethod
  def _is_present(val):
    """Returns True if the value is not None, and if it is either not a string, or a string with
    length > 0.
    """
    if val is None:
      return False
    if isinstance(val, str):
      return len(val) > 0
    return True

  @staticmethod
  def is_present(property_name):
    """Returns a Validation that returns false when the property is None or an empty string."""
    return Validation(Validation._is_present, property_name, "cannot be blank")

  @staticmethod
  def is_length(property_name, *, min_length=1, max_length=None, present_optional=False):
    """Returns a Validation that checks the length of a string."""
    def check(val):
      """Checks that a value matches a scope-enclosed set of length parameters."""
      if not val:
        return present_optional
      else:
        if len(val) >= min_length:
          if max_length is None:
            return True
          else:
            return len(val) <= max_length
        else:
          return False

    if max_length:
      message = "must be at least {0} characters long".format(min_length)
    else:
      message = "must be between {0} and {1} characters long".format(min_length, max_length)

    return Validation(check, property_name, message)

  @staticmethod
  def matches(property_name, regex, *, present_optional=False, message=None):
    """Returns a Validation that checks a property against a regex."""
    def check(val):
      """Checks that a value matches a scope-enclosed regex."""
      if not val:
        return present_optional
      else:
        return True if regex.search(val) else False

    return Validation(check, property_name, message)

  @staticmethod
  def is_numeric(property_name, *, numtype="float", min=None, max=None,
                 present_optional=False, message=None):
    """Returns a Validation that checks a property as a number, with optional range constraints."""
    if numtype == "int":
      cast = util.try_parse_int
    elif numtype == "decimal":
      cast = util.try_parse_decimal
    elif numtype == "float":
      cast = util.try_parse_float
    else:
      raise ValueError("numtype argument must be one of: int, decimal, float")

    def check(val):
      """Checks that a value can be parsed as a number."""
      if val is None:
        return present_optional
      else:
        is_num, new_val = cast(val)
        if not is_num:
          return False
        else:
          if min is not None and new_val < min:
            return False
          if max is not None and new_val > max:
            return False
          return True

    if not message:
      msg = ["must be a"]
      if numtype == "int":
        msg.append("whole number")
      else:
        msg.append("number")
      if min is not None and max is not None:
        msg.append("between {0} and {1}".format(min, max))
      elif min is not None:
        msg.append("greater than or equal to {0}".format(min))
      elif max is not None:
        msg.append("less than or equal to {0}".format(max))
      message = " ".join(msg)

    return Validation(check, property_name, message)

  @staticmethod
  def is_date(property_name, *, format=None, present_optional=False, message=None):
    """Returns a Validation that checks a value as a date."""
    # NOTE: Not currently using format param
    def check(val):
      """Checks that a value can be parsed as a date."""
      if val is None:
        return present_optional
      else:
        is_date, _ = util.try_parse_date(val)
        return is_date

    return Validation(check, property_name, message)

  @staticmethod
  def is_datetime(property_name, *, format=None, present_optional=False, message=None):
    """Returns a Validation that checks a value as a datetime."""
    # NOTE: Not currently using format param
    def check(val):
      """Checks that a value can be parsed as a datetime."""
      if val is None:
        return present_optional
      else:
        is_date, _ = util.try_parse_datetime(val)
        return is_date

    return Validation(check, property_name, message)

  @staticmethod
  def is_in(property_name, set_values, *, present_optional=False, message=None):
    """Returns a Validation that checks that a value is contained within a given set."""
    def check(val):
      """Checks that a value is contained within a scope-enclosed set."""
      if val is None:
        return present_optional
      else:
        return val in set_values

    return Validation(check, property_name, message)

  @staticmethod
  def is_unique(keys, *, scope=None, comparison_operators=None, present_optional=False,
                message=None):
    """Returns a Validation that makes sure the given value is unique for a table and optionally a
    scope.
    """
    def check(pname, validator):
      """Checks that a value is unique in its column, with an optional scope."""
      # pylint: disable=too-many-branches
      model = validator.model
      data_access = validator.data_access
      pkname = model.primary_key_name
      pkey = model.primary_key

      if isinstance(keys, str):
        key = getattr(model, keys)

        if present_optional and key is None:
          return True

        if comparison_operators:
          if isinstance(comparison_operators, str):
            op = comparison_operators
          else:
            op = comparison_operators[0]
        else:
          op = " = "

        constraints = [(keys, key, op)]
      else:
        if comparison_operators:
          ops = comparison_operators
        else:
          ops = [" = "] * len(keys)

        constraints = list(zip(keys, [getattr(model, key) for key in keys], ops))

      if scope:
        if comparison_operators:
          ops = comparison_operators[len(constraints):]
        else:
          ops = [" = "] * len(scope)

        constraints.extend(zip(scope, [getattr(model, col) for col in scope], ops))

      dupe = data_access.find(model.table_name, constraints, columns=pkname)

      if dupe is None:
        return True
      if isinstance(pkname, str):
        return dupe[0] == pkey
      else:
        return tuple(dupe) == tuple(pkey)

    return Validation(check, keys, message or "is already taken", is_simple=False)


class Validator(object):
  """The Validator class holds a collection of validations and methods to work with them."""
  def __init__(self, validations=None, *, fail_fast=False):
    """Initializes the Validator instance."""
    self.validations = validations or []
    self.model = None
    self.data_access = None
    self.fail_fast = fail_fast

  def add(self, validation):
    """Adds a Validation to the collection."""
    self.validations.append(validation)

  def validate(self, model, data_access=None, *, fail_fast=None):
    """Validates a model against the collection of Validations.

    Returns True if all Validations pass, or False if one or more do not.
    """
    if fail_fast is None:
      fail_fast = self.fail_fast

    self.model = model
    self.data_access = data_access
    is_valid = True
    for validation in self.validations:
      if not validation.validate(model, self):
        is_valid = False
        if fail_fast:
          break
    return is_valid



import unittest
from decimal import Decimal
from uuid import UUID, uuid4
from datetime import datetime, timezone, time

import bidon.util.convert as cv


class UtilConvertTestCase(unittest.TestCase):
  def test_identity(self):
    self.assertEqual(cv.identity(5), 5)
    self.assertEqual(cv.identity("18"), "18")
    self.assertEqual(cv.identity(None), None)
    self.assertEqual(cv.identity([1,2,3]), [1,2,3])

  def test_to_int(self):
    self.assertEqual(cv.to_int(3), 3)
    self.assertEqual(cv.to_int("3"), 3)
    self.assertRaises(Exception, lambda: cv.to_int("3.3"))
    self.assertRaises(Exception, lambda: cv.to_int(None))

  def test_to_float(self):
    self.assertEqual(cv.to_float(3.3), 3.3)
    self.assertEqual(cv.to_float("3.3"), 3.3)
    self.assertRaises(Exception, lambda: cv.to_float("3.a3"))
    self.assertRaises(Exception, lambda: cv.to_float(None))

  def test_to_decimal(self):
    self.assertEqual(cv.to_decimal(Decimal(3.4)), Decimal(3.4))
    self.assertEqual(cv.to_decimal(3), Decimal(3))
    self.assertEqual(cv.to_decimal(3.4), Decimal(3.4))
    self.assertEqual(cv.to_decimal("3.4"), Decimal("3.4"))
    self.assertRaises(Exception, lambda: cv.to_decimal("a"))
    self.assertRaises(Exception, lambda: cv.to_decimal(None))

  def test_to_bit(self):
    self.assertEqual(cv.to_bit(0), 0)
    self.assertEqual(cv.to_bit(False), 0)
    self.assertEqual(cv.to_bit(1), 1)
    self.assertEqual(cv.to_bit(True), 1)

  def test_to_bool(self):
    self.assertEqual(cv.to_bit(0), False)
    self.assertEqual(cv.to_bit(False), False)
    self.assertEqual(cv.to_bit(1), True)
    self.assertEqual(cv.to_bit(True), True)

  def test_to_uuid(self):
    u = uuid4()
    self.assertEqual(cv.to_uuid(str(u)), u)
    self.assertEqual(cv.to_uuid(u), u)
    self.assertRaises(Exception, lambda: cv.to_uuid(str(u)[:-1]))
    self.assertRaises(Exception, lambda: cv.to_uuid(None))

  def test_to_compressed_string(self):
    tests = [
      (None, None),
      ("", None),
      ("   ", None),
      ("a", "a"),
      (" a", "a"),
      ("a ", "a"),
      (" a  ", "a"),
      ("a b", "a b"),
      ("a  b c", "a b c"),
      ("  a  b   c d    ", "a b c d")
    ]
    for i, o in tests:
      self.assertEqual(cv.to_compressed_string(i), o)
    self.assertEqual(cv.to_compressed_string("  a  b cd ", 5), "a b c")


  def test_to_now(self):
    start = datetime.now().timestamp()

    self.assertLessEqual(start, cv.to_now(None).timestamp())
    self.assertLessEqual(cv.to_now(None).timestamp(), datetime.now().timestamp())

  def test_to_none(self):
    self.assertIsNone(cv.to_none("ABC"))
    self.assertIsNone(cv.to_none(123))
    self.assertIsNone(cv.to_none(None))

  def test_to_true(self):
    self.assertTrue(cv.to_true("ABC"))
    self.assertTrue(cv.to_true(123))
    self.assertTrue(cv.to_true(None))

  def test_to_false(self):
    self.assertFalse(cv.to_false("ABC"))
    self.assertFalse(cv.to_false(123))
    self.assertFalse(cv.to_false(None))

  def test_to_empty_string(self):
    self.assertEqual(cv.to_empty_string("ABC"), "")
    self.assertEqual(cv.to_empty_string(123), "")
    self.assertEqual(cv.to_empty_string(None), "")

  def test_to_new_uuid(self):
    self.assertIsInstance(cv.to_new_uuid(None), UUID)

  def test_incrementor(self):
    i0 = cv.incrementor()
    self.assertEqual(i0(None), 0)
    self.assertEqual(i0(None), 1)
    i1 = cv.incrementor(start=1, step=10)
    self.assertEqual(i1(None), 1)
    self.assertEqual(i1(None), 11)
    self.assertEqual(i1(None), 21)

  def test_string_trimmer(self):
    st0 = cv.string_trimmer()
    st1 = cv.string_trimmer(5)
    self.assertEqual(st0(" a"), "a")
    self.assertEqual(st0("a "), "a")
    self.assertEqual(st0(" a "), "a")
    self.assertEqual(st1("  abcdef  "), "abcde")

  def test_staic_value(self):
    svs = [None, True, 1, {}]
    inps = ["a", 3, []]
    for sv in svs:
      svfx = cv.static_value(sv)
      for inp in inps:
        self.assertEqual(svfx(inp), sv)

  def test_rounded_decimal(self):
    n = Decimal('3.926')
    conv = cv.rounded_decimal(2)
    self.assertEqual(conv(n), Decimal('3.93'))

  def test_index_resolver(self):
    index = dict(a=1, b=2)
    r1 = cv.index_resolver(index, False)
    r2 = cv.index_resolver(index, True)

    self.assertEqual(r1("a"), 1)
    self.assertEqual(r2("b"), 2)
    self.assertEqual(r1("c"), None)
    with self.assertRaises(KeyError):
      r2("d")

  def test_accept_none_wrapper(self):
    conv = cv.accept_none_wrapper(cv.to_int)
    self.assertEqual(conv(None), None)
    self.assertEqual(conv("3"), 3)
    self.assertEqual(conv(3), 3)

  def test_try_wrapper(self):
    conv = cv.try_wrapper(cv.to_int)
    self.assertEqual(conv("3"), (True, 3))
    self.assertEqual(conv(3), (True, 3))
    self.assertFalse(conv("3.5")[0])



import json
import unittest
from datetime import datetime
from collections import namedtuple

import bidon.util as h

from .test_convert import UtilConvertTestCase
from .test_date import UtilDateTestCase
from .test_transform import UtilTransformTestCase


class UtilTestCase(unittest.TestCase):
  def test_pick(self):
    d = dict(a=1, b=2, c=3, d=4)
    self.assertEqual(h.pick(d, ["a", "b"]), dict(a=1, b=2))
    self.assertEqual(h.pick(d, lambda k: k <= "c"), dict(a=1, b=2, c=3))
    self.assertEqual(h.pick(d, ["a", "b"], transform=lambda n: n * 2), dict(a=2, b=4))

  def test_exclude(self):
    d = dict(a=1, b=2, c=3, d=4)
    self.assertEqual(h.exclude(d, ["a", "b"]), dict(c=3, d=4))
    self.assertEqual(h.exclude(d, lambda k: k >= "c"), dict(a=1, b=2))
    self.assertEqual(h.exclude(d, ["a", "b"], transform=lambda n: n * 2), dict(c=6, d=8))

  def test_to_json(self):
    d = dict(a=[1, 2])
    self.assertEqual(h.to_json(d), json.dumps(d, sort_keys=False, indent=None, separators=(",", ":")))
    self.assertEqual(h.to_json(d, True), json.dumps(d, sort_keys=True, indent=2, separators=(", ", ": ")))
    with self.assertRaises(TypeError):
      d = dict(dt=datetime.now())
      h.to_json(d)
    h.register_json_default(lambda o: hasattr(o, "isoformat"), lambda o: o.isoformat())
    self.assertEqual(h.to_json(d), h.to_json(h.pick(d, ["dt"], transform=lambda n: n.isoformat())))

  def test_has_value(self):
    class A(object):
      def __init__(self):
        self.n = 10

    self.assertEqual(h.has_value(A(), "n"), (True, 10))
    self.assertEqual(h.has_value(A(), "d"), (False, None))
    self.assertEqual(h.has_value(dict(n=10), "n"), (True, 10))
    self.assertEqual(h.has_value(dict(v=10), "n"), (False, None))
    self.assertEqual(h.has_value(None, "n"), (False, None))

  def test_get_value(self):
    class A(object):
      def __init__(self):
        self.n = 10

    self.assertEqual(h.get_value(A(), "n"), 10)
    self.assertEqual(h.get_value(A(), "d"), None)
    self.assertEqual(h.get_value(dict(n=10), "n"), 10)
    self.assertEqual(h.get_value(dict(v=10), "n", "hi"), "hi")
    self.assertEqual(h.get_value(dict(v=10), "n", lambda: "lo"), "lo")
    self.assertEqual(h.get_value(None, "n", 42), 42)

  def test_set_value(self):
    class A(object):
      def __init__(self):
        self.n = 10

    a = A()
    b = {}
    c = []
    h.set_value(a, "n", 15)
    self.assertEqual(a.n, 15)
    h.set_value(b, "v", 18)
    self.assertEqual(b["v"], 18)
    self.assertRaises(Exception, lambda: h.set_value(c, "q", 13))

  def test_with_defaults(self):
    def a(b,c,d): return [b,c,d]
    self.assertEqual([None, None, None], h.with_defaults(a, 3))
    self.assertEqual([1, None, None], h.with_defaults(a, 3, [1]))
    self.assertEqual([1, 2, 3], h.with_defaults(a, 3, [1, 2, 3]))

  def test_namedtuple_with_defaults(self):
    nt = namedtuple("Point3d", ["x", "y", "z"])
    self.assertEqual(nt(None, None, None), h.namedtuple_with_defaults(nt))
    self.assertEqual(nt(1, None, None), h.namedtuple_with_defaults(nt, [1]))
    self.assertEqual(nt(1, 2, 3), h.namedtuple_with_defaults(nt, [1, 2, 3]))

  def test_flatten_dict(self):
    self.assertEqual(
      sorted(h.flatten_dict(
        {"a": {"c": 1, "d": 2},
         "b": {"c": 10, "d": 11, "e": 12},
         "c": 20})),
      [(("a", "c"), 1),
       (("a", "d"), 2),
       (("b", "c"), 10),
       (("b", "d"), 11),
       (("b", "e"), 12),
       (("c", ), 20)])



"""This module implements a wrapper around the standard library date and time functions to simplify
the parsing of ISO8601 date strings.
"""
import itertools
import re
from datetime import datetime, date, time, timezone


def local_now():
  """Returns a datetime object for the current time, timezone unaware."""
  return datetime.now()


def utc_now():
  """Returns a datetime object for the current time, with the tzinfo set to timezone.utc"""
  return datetime.utcnow().replace(tzinfo=timezone.utc)


def parse_date(val, fmt=None):
  """Returns a date object parsed from :val:.

  :param val: a string to be parsed as a date
  :param fmt: a format string, a tuple of format strings, or None. If None a
              built in list of format strings will be used to try to parse the
              date.
  """
  if isinstance(val, date):
    return val
  else:
    return parse_datetime(val, fmt).date()


def parse_time(val, fmt=None):
  """Returns a time object parsed from :val:.

  :param val: a string to be parsed as a time
  :param fmt: a format string, a tuple of format strings, or None. If None a built in list of format
              strings will be used to try to parse the time.
  """
  if isinstance(val, time):
    return val
  else:
    return parse_datetime(val, fmt).time()


def parse_datetime(val, fmt=None):
  """Returns a datetime object parsed from :val:.

  :param val: a string to be parsed as a date
  :param fmt: a format string, a tuple of format strings, or None. If None a built in list of format
              strings will be used to try to parse the date.
  """
  if isinstance(val, datetime):
    return val
  else:
    return _parse_datetime(_normalize_tz(str(val)), fmt)


def _parse_datetime(val, fmt=None):
  """Returns a datetime object parsed from :val:. The timezone, if any, on the
  string must be in the format that python can parse.

  :param val: a string to be parsed as a date
  :param fmt: a format string, a tuple of format strings, or None. If None a built in list of format
              strings will be used to try to parse the datetime.
  """
  if fmt is None:
    fmt = _DEFAULT_FORMATS

  if isinstance(fmt, str):
    return datetime.strptime(val, fmt)
  else:
    for opt in fmt:
      try:
        return _parse_datetime(val, opt)
      except ValueError:
        continue
    raise ValueError("Unable to parse string '{}' as a timestamp".format(val))


def _normalize_tz(val):
  """Normalizes all valid ISO8601 time zone variants to the one python will
  parse.

  :val: a timestamp string without a timezone, or with a timezone in one of the ISO8601 accepted
        formats.
  """
  match = _TZ_RE.match(val)
  if match:
    ts, tz = match.groups()
    if len(tz) == 5:
      # If the length of the tz is 5 then it is of the form (+|-)dddd, which is exactly what python
      # wants, so just return it.
      return ts + tz
    if len(tz) == 6:
      # If the length of the tz is 6 then it is of the form (+|-)dd:dd, just remove the colon
      return ts + tz[:3] + tz[4:]
    if tz == "Z" or tz == "z":
      # If the tz is "Z" or 'z', return a timezone of +0000
      return ts + "+0000"
    else:
      # Otherwise, the timzone must be of the format (+|-)dd, in which case we just need to add two
      # "0" to it, and it will be in the proper format.
      return ts + tz + "00"
  else:
    return val


def _join_date_and_time(dates, times, joiner):
  """Returns a tuple of all date and time format strings joined together by
  :joiner:.

  :dates: an enumerable of date format strings
  :times: an enumerable of time format strings
  :joiner: a string to join a date and time format together
  """
  return tuple("{}{}{}".format(d, joiner, t) for (d, t) in itertools.product(dates, times))


# A timezone can be one of:
# (+|-)dd, (+|-)dddd, (+|-)dd:dd, Z
_TZ_RE = re.compile(r"^(.*)((?:\+|-)\d\d(?::?\d\d)?|Z)$", re.IGNORECASE)

_DATE_FORMATS = (
  "%Y-%m-%d",
  "%Y%m%d")

_TIME_FORMATS_24_DELIM = (
  "%H:%M",
  "%H:%M:%S",
  "%H:%M:%S.%f")

_TIME_FORMATS_24_NODELIM = (
  "%H%M",
  "%H%M%S",
  "%H%M%S.%f")

_TIME_FORMATS_12 = (
  "%I:%M %p",
  "%I:%M:%S %p",
  "%I:%M:%S.%f %p")

_TIME_FORMATS = _TIME_FORMATS_24_DELIM \
                + _TIME_FORMATS_24_NODELIM \
                + _TIME_FORMATS_12 \
                + _join_date_and_time(_DATE_FORMATS, _TIME_FORMATS_24_DELIM, " ") \
                + _join_date_and_time(_DATE_FORMATS, _TIME_FORMATS_24_DELIM, "T") \
                + _join_date_and_time(_DATE_FORMATS, _TIME_FORMATS_24_NODELIM, "") \
                + _join_date_and_time(_DATE_FORMATS, _TIME_FORMATS_24_NODELIM, "T")

_DEFAULT_FORMATS = _DATE_FORMATS \
                   + _TIME_FORMATS \
                   + tuple("{}%z".format(f) for f in _TIME_FORMATS)



from setuptools import setup, find_packages

version = "1.0.3"

setup(name="bidon",
      version=version,
      description="A simple, easy to use, and flexible data handling library",
      author="Trey Cucco",
      author_email="fcucco@gmail.com",
      url="https://github.com/treycucco/bidon",
      download_url="https://github.com/treycucco/bidon/tarball/master",
      packages=["bidon",
                "bidon.db",
                "bidon.db.access",
                "bidon.db.core",
                "bidon.db.model",
                "bidon.experimental",
                "bidon.spreadsheet",
                "bidon.util",
                "bidon.xml"],
      classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3 :: Only"],
      license="BSD",
      platforms="any")



"""This module contains methods for working with Postgres advisory locks."""
from contextlib import contextmanager
from enum import Enum


__all__ = ["LockMode", "lock_key", "obtain_lock", "release_lock", "advisory_lock"]


class LockMode(Enum):
  """Contains the options for how to perform Postgres advisory locks."""
  wait = 0x01
  skip = 0x02
  error = 0x04


def lock_key(group_id, item_id, group_width=8):
  """Creates a lock ID where the lower bits are the group ID and the upper bits
  are the item ID. This allows the use of a bigint namespace for items, with a
  limited space for grouping.

  :group_id: an integer identifying the group. Must be less than
             2 ^ :group_width:
  :item_id: item_id an integer. must be less than 2 ^ (63 - :group_width:) - 1
  :gropu_width: the number of bits to reserve for the group ID.
  """
  if group_id >= (1 << group_width):
    raise Exception("Group ID is too big")
  if item_id >= (1 << (63 - group_width)) - 1:
    raise Exception("Item ID is too big")
  return (item_id << group_width) | group_id


def obtain_lock(dax, key, lock_mode=LockMode.wait, xact=False):
  """Runs sql to obtain a pg advisory lock.

  :dax: a DataAccess instance
  :key: either a big int or a 2-tuple of integers
  :lock_mode: a member of the LockMode enum
  :xact: a boolean, if True the lock will be automatically released at the end
         of the transaction and cannot be manually released.
  """
  lock_fxn = _lock_fxn("lock", lock_mode, xact)
  return dax.get_scalar(
    dax.callproc(lock_fxn, key if isinstance(key, (list, tuple)) else [key])[0])


def release_lock(dax, key, lock_mode=LockMode.wait):
  """Manually release a pg advisory lock.

  :dax: a DataAccess instance
  :key: either a big int or a 2-tuple of integers
  :lock_mode: a member of the LockMode enum
  """
  lock_fxn = _lock_fxn("unlock", lock_mode, False)
  return dax.get_scalar(
    dax.callproc(lock_fxn, key if isinstance(key, (list, tuple)) else [key])[0])


@contextmanager
def advisory_lock(dax, key, lock_mode=LockMode.wait, xact=False):
  """A context manager for obtaining a lock, executing code, and then releasing
  the lock.

  A boolean value is passed to the block indicating whether or not the lock was
  obtained.

  :dax: a DataAccess instance
  :key: either a big int or a 2-tuple of integers
  :lock_mode: a member of the LockMode enum. Determines how this function
              operates:
                - wait: the wrapped code will not be executed until the lock
                        is obtained.
                - skip: an attempt will be made to get the lock, and if
                        unsuccessful, False is passed to the code block
                - error: an attempt will be made to get the lock, and if
                         unsuccessful, an exception will be raised.
  :xact: a boolean, if True, the lock will be obtained according to lock_mode,
         but will not be released after the code is executed, since it will be
         automatically released at the end of the transaction.
  """
  if lock_mode == LockMode.wait:
    obtain_lock(dax, key, lock_mode, xact)
  else:
    got_lock = obtain_lock(dax, key, lock_mode, xact)
    if not got_lock:
      if lock_mode == LockMode.error:
        raise Exception("Unable to obtain advisory lock {}".format(key))
      else:
        # lock_mode is skip
        yield False
        return

  # At this point we have the lock
  try:
    yield True
  finally:
    if not xact:
      release_lock(dax, key, lock_mode)


def _lock_fxn(direction, lock_mode, xact):
  """Builds a pg advisory lock function name based on various options.

  :direction: one of "lock" or "unlock"
  :lock_mode: a member of the LockMode enum
  :xact: a boolean, if True the lock will be automatically released at the end
         of the transaction and cannot be manually released.
  """
  if direction == "unlock" or lock_mode == LockMode.wait:
    try_mode = ""
  else:
    try_mode = "_try"

  if direction == "lock" and xact:
    xact_mode = "_xact"
  else:
    xact_mode = ""

  return "pg{}_advisory{}_{}".format(try_mode, xact_mode, direction)



"""This module contains classes that are used to compose data access objects."""
from .data_access_core import DataAccessCoreBase, InjectedDataAccessCore, get_pg_core
from .data_access_core import get_pooled_pg_core, get_sqlite_core, get_mysql_core
from .sql_writer import SqlWriter



"""The model_base module contains the ModelBase class."""
from bidon.util import get_value, to_json, pick, exclude
from .validation import Validator


__all__ = ["ModelBase"]


class ModelBase(object):
  """A simple model base class allowing easy specification and creation of database backed objects.
  """
  table_name = None
  primary_key_name = "id"
  primary_key_is_auto = True
  timestamps = ("created_at", "updated_at")
  attrs = {}
  validator = Validator()
  exclude_keys_serialize = set()
  exclude_keys_sql = {"created_at", "updated_at"}
  strict_attrs = False

  def __init__(self, data_=None, **attrs_):
    """Initializes the ModelBase instance.

    :data_: can be either a dictionary, another Model, or None. The instance is initialized from it.
    :attrs: are passed to the instance's update method after the instance is initialized from data_.
    """
    self._initialize_attributes(data_ or {})
    self.update(attrs_)
    self.errors = {}

  @classmethod
  def create_model_class(cls, name, table_name, attrs, validations=None, *, otherattrs=None,
                         other_bases=None):
    """Creates a new class derived from ModelBase."""
    members = dict(table_name=table_name, attrs=attrs, validations=validations or [])
    if otherattrs:
      members.update(otherattrs)
    return type(name, other_bases or (), members)

  @classmethod
  def has_attr(cls, attr_name):
    """Check to see if an attribute is defined for the model."""
    if attr_name in cls.attrs:
      return True

    if isinstance(cls.primary_key_name, str) and cls.primary_key_name == attr_name:
      return True

    if isinstance(cls.primary_key_name, tuple) and attr_name in cls.primary_key_name:
      return True

    if cls.timestamps is not None and attr_name in cls.timestamps:
      return True

    return False

  @property
  def primary_key(self):
    """Returns either the primary key value, or a tuple containing the primary key values in the
    case of a composite primary key.
    """
    pkname = self.primary_key_name
    if pkname is None:
      return None
    elif isinstance(pkname, str):
      return getattr(self, pkname)
    else:
      return tuple((getattr(self, pkn) for pkn in pkname))

  @property
  def is_new(self):
    """Determines whether or not a model has been saved to the database or not by checking its
    primary key value.
    """
    if not self.primary_key_is_auto:
      raise Exception("Unable to determine newness of non-auto increment primary key moodels")
    return self.primary_key is None

  def before_validation(self):
    """Method to call before validation is run. Allows subclasses to fixup their properties."""
    pass

  def validate(self, data_access=None):
    """Run the class validations against the instance. If the validations require database access,
    pass in a DataAccess derived instance.
    """
    self.clear_errors()
    self.before_validation()
    self.validator.validate(self, data_access)
    return not self.has_errors

  @property
  def has_errors(self):
    """Returns True if there are any errors recorded for the instance."""
    return bool(self.errors)

  def add_error(self, property_name, message):
    """Add an error for the given property."""
    if property_name not in self.errors:
      self.errors[property_name] = []
    self.errors[property_name].append(message)

  def clear_errors(self):
    """Clears all errors."""
    self.errors.clear()

  def all_errors(self, joiner="; "):
    """Returns a string representation of all errors recorded for the instance."""
    parts = []
    for pname, errs in self.errors.items():
      for err in errs:
        parts.append("{0}: {1}".format(pname, err))
    return joiner.join(parts)

  def to_dict(self, *, include_keys=None, exclude_keys=None, use_default_excludes=True):
    """Converts the class to a dictionary.

    :include_keys: if not None, only the attrs given will be included.
    :exclude_keys: if not None, all attrs except those listed will be included, with respect to
    use_default_excludes.
    :use_default_excludes: if True, then the class-level exclude_keys_serialize will be combined
    with exclude_keys if given, or used in place of exlcude_keys if not given.
    """
    data = self.__dict__

    if include_keys:
      return pick(data, include_keys, transform=self._other_to_dict)
    else:
      skeys = self.exclude_keys_serialize if use_default_excludes else None
      ekeys = exclude_keys

      return exclude(
        data,
        lambda k: (skeys is not None and k in skeys) or (ekeys is not None and k in ekeys),
        transform=self._other_to_dict)

  def to_json(self, *, include_keys=None, exclude_keys=None, use_default_excludes=True,
              pretty=False):
    """Converts the response from to_dict to a JSON string. If pretty is True then newlines,
    indentation and key sorting are used.
    """
    return to_json(
      self.to_dict(
        include_keys=include_keys,
        exclude_keys=exclude_keys,
        use_default_excludes=use_default_excludes),
      pretty=pretty)

  def update(self, data_=None, **kwargs):
    """Update the object with the given data object, and with any other key-value args. Returns a
    set containing all the property names that were changed.
    """
    if data_ is None:
      data_ = dict()
    else:
      data_ = dict(data_)
    data_.update(**kwargs)
    changes = set()
    for attr_name in data_:
      if hasattr(self, attr_name):
        if getattr(self, attr_name) != data_[attr_name]:
          changes.add(attr_name)
          setattr(self, attr_name, data_[attr_name])
      else:
        if self.strict_attrs:
          raise Exception("Unknown attribute for {}: {}".format(self.__class__.__name__, attr_name))
    return changes

  def _initialize_attributes(self, data):
    """Initializes the class with the given data. Used by __init__."""
    attrs = dict(self.attrs)

    if self.primary_key_name is None:
      pass
    elif isinstance(self.primary_key_name, str):
      if self.primary_key_name not in attrs:
        attrs[self.primary_key_name] = None
    else:
      for pkey in self.primary_key_name:
        if pkey not in attrs:
          attrs[pkey] = None

    if self.timestamps:
      for ts_name in self.timestamps:
        if ts_name is not None and ts_name not in attrs:
          attrs[ts_name] = None

    for attr_name in attrs:
      setattr(self, attr_name, get_value(data, attr_name, attrs[attr_name]))

  def _other_to_dict(self, other):
    """When serializing models, this allows attached models (children, parents, etc.) to also be
    serialized.
    """
    if isinstance(other, ModelBase):
      return other.to_dict()
    elif isinstance(other, list):
      # TODO: what if it's not a list?
      return [self._other_to_dict(i) for i in other]
    else:
      return other
